--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the crypto_int_arbiter module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- @date 2015-01-26

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

--library des_lib;
--use des_lib.des_pkg.all;

entity crypto_int_arbiter_sim is

generic(
      n_t         : natural := 100000;      -- Number of test vectors
      rnd_ce      : boolean := false;
      debug       : boolean := false
);

end entity crypto_int_arbiter_sim;

architecture sim of crypto_int_arbiter_sim is
constant NRQ : natural := 2;

signal clk, srstn, ce          : std_ulogic;
signal eos, eof                : boolean;

signal s_rd_crypto_cmd_a       : crypto_cmd_out_array(NRQ - 1 downto 0);
signal s_wr_crypto_cmd_a       : crypto_cmd_in_array(NRQ  - 1 downto 0);  -- cmdack, rspval
signal s_rd_crypto_data_ms     : data_vector(11 downto 0); -- Data set = 6 blocks 
signal s_rd_crypto_data_mt     : data_vector(11 downto 0);
signal s_wr_crypto_cmd         : crypto_cmd_out;  
signal s_rd_crypto_cmd         : crypto_cmd_in;  -- cmdack, rspval
signal s_wr_crypto_data        : data_vector(11 downto 0);

signal index : natural;

begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
  variable l         : line   ;

  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

    for i in 1 to n_t loop
      if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
      end if;
      
      wait until rising_edge(clk) and ce = '1';

    end loop;

    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

icryptoIntArb : entity work.crypto_int_arbiter(rtl) 
generic map(
       NRQ               => NRQ
)
port map(
       clk               => clk                 , 
       srstn             => srstn               , 
       ce                => ce                  , 
                                
       rd_crypto_cmd_a   => s_rd_crypto_cmd_a   ,
       rd_crypto_data_ms => s_rd_crypto_data_ms ,
       rd_crypto_data_mt => s_rd_crypto_data_mt ,
       rd_crypto_cmd     => s_rd_crypto_cmd     ,
       wr_crypto_cmd     => s_wr_crypto_cmd     ,
       wr_crypto_cmd_a   => s_wr_crypto_cmd_a   ,
       wr_crypto_data    => s_wr_crypto_data    
);

stim : process
variable k   : natural;
variable val : std_ulogic_vector(1 downto 0);
begin

   for i in 0 to NRQ - 1 loop
      s_rd_crypto_cmd_a(i).cmdval      <= '0'; 
      s_rd_crypto_cmd_a(i).no_init     <= '0'; 
      s_rd_crypto_cmd_a(i).direct_data <= '0'; 
      s_rd_crypto_cmd_a(i).mode        <= CRY_MAC;
      s_rd_crypto_cmd_a(i).key         <= key_null;
      s_rd_crypto_cmd_a(i).iv          <= (others => '0');  -- not used
      s_rd_crypto_cmd_a(i).size        <= (others => '0');      
      s_rd_crypto_cmd_a(i).cipher      <= '0';
      s_rd_crypto_cmd_a(i).id_src      <= (others => '0');
      s_rd_crypto_cmd_a(i).id_dest     <= (others => '0');
      s_rd_crypto_cmd_a(i).address     <= (others => '0');
   end loop;

   s_rd_crypto_data_ms <= (others => (others => '0'));
   s_rd_crypto_data_mt <= (others => (others => '0'));

   for i in 30  to 50 loop
       wait until rising_edge(clk) and ce='1';
   end loop;

   while not eos loop
         for i in 30  to 50 loop
             wait until rising_edge(clk) and ce='1';
         end loop;
  
         val := std_ulogic_vector_rnd(val'length);
         for i in 0 to NRQ - 1 loop
             s_rd_crypto_cmd_a(i).cmdval    <= val(i);    
             s_rd_crypto_cmd_a(i).address   <= std_ulogic_vector_rnd(ADDR_SIZE); 
         end loop;

         if val(0) = '1' or val(1) = '1' then 
             wait until rising_edge(clk) and ce='1' and (s_wr_crypto_cmd_a(0).cmdack='1' or s_wr_crypto_cmd_a(1).cmdack='1');
             assert (s_wr_crypto_cmd_a(index).cmdack = '1')  
             report "Wrong priority cmdack output index ... "
             severity Failure;

             for i in 0 to NRQ - 1 loop
                 s_rd_crypto_cmd_a(i).cmdval    <= '0';   
                 s_rd_crypto_cmd_a(i).address   <= (others => '0'); 
             end loop;

             wait until rising_edge(clk) and ce='1' and (s_wr_crypto_cmd_a(0).rspval='1' or s_wr_crypto_cmd_a(1).rspval='1');
             assert (s_wr_crypto_cmd_a(index).rspval = '1')  
             report "Wrong priority rspval output index ... "
             severity Failure;
          end if;
   end loop;

end process;

verif: process
variable tmps  : natural;
begin
    s_rd_crypto_cmd.cmdack <= '0';
    s_rd_crypto_cmd.rspval <= '0';

    -- IDLE
    wait until rising_edge(clk) and ce='1' and (s_rd_crypto_cmd_a(0).cmdval='1' or s_rd_crypto_cmd_a(1).cmdval='1');
    if s_rd_crypto_cmd_a(0).cmdval='1' then 
       index <= 0;
    else 
       index <= 1;
    end if;

    wait until rising_edge(clk) and ce='1'; 
    s_rd_crypto_cmd.cmdack <= '1';

    tmps := integer_rnd(10,20);
    for i in 1 to tmps loop 
          wait until rising_edge(clk) and ce = '1';
    end loop;
 
    s_rd_crypto_cmd.rspval <= '1';
    s_rd_crypto_cmd.cmdack <= '0';

    wait until rising_edge(clk) and ce='1'; 
    s_rd_crypto_cmd.rspval <= '0';

end process;


 
end architecture sim;
