--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the modules  axi_2vci and vci_2_axi
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- @date 2015-02-16

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library axi_lib;
use axi_lib.axi_pkg.all;

library vci_ram_lib;
use vci_ram_lib.all;

entity axi_vci_sim is

  generic(n_t: positive := 100000;      -- Number of test vectors
          rnd_ce: boolean := false;
          debug: boolean := false);
end entity axi_vci_sim;

architecture sim of axi_vci_sim is

  signal clk, srstn, ce: std_ulogic;
  signal eos: boolean;

  signal s_vci_i2t     : vci_i2t;
  signal s_vci_t2i     : vci_t2i;
  signal s_axi_m2s_in  : axi_gp_m2s;
  signal s_axi_s2m_out : axi_gp_s2m;
  signal s_axi_m2s_out : axi_gp_m2s; 
  signal s_axi_s2m_in  : axi_gp_s2m;

  signal s_addr        : addr_t;
  signal s_id          : srcid_t;
  signal s_data        : data_vector(15 downto 0);
  signal s_be          : be_vector(15 downto 0);
  signal s_len         : unsigned(AXI_L - 1 downto 0);   

begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l        : line;
 
  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

   for i in 1 to n_t  loop
     if (i * 10) mod n_t = 0 then
         write(l, i);
         write(l, string'("/"));
         write(l, n_t);
         writeline(output, l);
     end if;
     wait until rising_edge(clk) and ce='1'; 

   end loop;

   eos <= true;
   write(l, string'("*** regression test passed"));
   writeline(output, l);
   wait;

  end process;

ivci_2_axi : entity work.vci_2_axi4(rtl) 
port map(
       clk         => clk           , 
       srstn       => srstn         , 
       ce          => ce            , 

       -- Axi4 out 
       p_axi_in    => s_axi_s2m_in  , -- in   
       p_axi_out   => s_axi_m2s_out , -- out  

       -- Vci in  
       p_vci_out   => s_vci_t2i     , -- out 
       p_vci_in    => s_vci_i2t       -- in  

);

iaxi4_2_vci : entity work.axi4_2_vci(rtl)
port map(
       clk         => clk           , 
       srstn       => srstn         , 
       ce          => ce            ,

       -- Axi4 in
       p_axi_in    => s_axi_m2s_in  , -- in   
       p_axi_out   => s_axi_s2m_out , -- out  

       -- Vci out 
       p_vci_out   => s_vci_i2t     , -- out 
       p_vci_in    => s_vci_t2i       -- in  
);


iaxi_ram: entity vci_ram_lib.axi_ram(rtl)
  port map(
           clk        => clk,
           srstn      => srstn,
           ce         => ce,

           p_axi_in   => s_axi_m2s_out ,
           p_axi_out  => s_axi_s2m_in  
);
 
  stim : process
  variable tmps  : natural;
  variable v_len : natural;                            

  begin

    while not eos loop
            s_axi_m2s_in <= axi_gp_m2s_none; 
 
            tmps := integer_rnd(10,50);
            for i in 1 to tmps loop 
                  wait until rising_edge(clk) and ce = '1';
            end loop;          

            s_id   <= std_ulogic_vector_rnd(SRCID_SIZE);
         -- s_addr <= (others => '0');
         -- s_addr(15 downto 2) <= std_ulogic_vector_rnd(14);
            s_addr <= std_ulogic_vector_rnd(ADDR_SIZE);
            
            s_len <= unsigned_rnd(AXI_L) - 1;

            for i in 0 to 15 loop
                 s_data(i) <= std_ulogic_vector_rnd(DATA_SIZE);
                 s_be(i)   <= (others => '1');
            end loop; 

            tmps := integer_rnd(10,50);
            for i in 1 to tmps loop 
                  wait until rising_edge(clk) and ce = '1';
            end loop;          

            
            -- send AXI ADDR WRITE REQ 
            s_axi_m2s_in.awid      <= s_id;
            s_axi_m2s_in.awaddr    <= s_addr; 
            s_axi_m2s_in.awlen     <= std_ulogic_vector(s_len);
            s_axi_m2s_in.awsize    <= b"010"; -- 4 bytes
            s_axi_m2s_in.awburst   <= b"01";  -- incr
            s_axi_m2s_in.awvalid   <= '1';

            wait until rising_edge(clk) and ce = '1' and s_axi_s2m_out.awready = '1';
            s_axi_m2s_in <= axi_gp_m2s_none;

            -- send AXI DATA WRITE BURST
            v_len := to_integer(s_len);

            for i in 0 to v_len loop 
                s_axi_m2s_in.wdata     <= s_data(i);
                s_axi_m2s_in.wstrb     <= s_be(i); 
                s_axi_m2s_in.wvalid    <= '1';

                if i = v_len  then
                   s_axi_m2s_in.wlast  <= '1';
                else
                   s_axi_m2s_in.wlast  <= '0';
                end if;
 
                wait until rising_edge(clk) and ce = '1' and s_axi_s2m_out.wready = '1';
            end loop;

            s_axi_m2s_in <= axi_gp_m2s_none; 
            s_axi_m2s_in.bready <= '1';   -- rspack

            wait until rising_edge(clk) and ce = '1' and s_axi_s2m_out.bvalid = '1';
            assert (s_axi_s2m_out.bid = s_id)
            report "bid mismatch !!!"  
            severity failure;

            s_axi_m2s_in.bready <= '0';   -- rspack

            tmps := integer_rnd(10,50);
            for i in 1 to tmps loop 
                  wait until rising_edge(clk) and ce = '1';
            end loop; 

            -- send AXI READ REQ
            s_axi_m2s_in.arid      <= s_id; 
            s_axi_m2s_in.araddr    <= s_addr; 
            s_axi_m2s_in.arlen     <= std_ulogic_vector(s_len);
            s_axi_m2s_in.arsize    <= b"010"; -- 4 bytes
            s_axi_m2s_in.arburst   <= b"01";  -- incr
            s_axi_m2s_in.arvalid   <= '1';
             
            wait until rising_edge(clk) and ce = '1' and s_axi_s2m_out.arready = '1';
            s_axi_m2s_in <= axi_gp_m2s_none; 
            s_axi_m2s_in.rready <= '1';   -- rspack

--          wait until rising_edge(clk) and ce = '1' and s_axi_s2m_out.rvalid = '1' and s_axi_s2m_out.rlast = '1';

            -- Check Matching data

            v_len := to_integer(s_len);
          for  i in 0 to v_len loop       
              wait until rising_edge(clk) and ce = '1' and s_axi_s2m_out.rvalid = '1';
             --if s_axi_s2m_out.rdata = s_data(i) then
             --    write(l, string'("  index : "));
             --    write(l, i);
             --    write(l, string'("  axi_data : "));
             --   hwrite(l, s_axi_s2m_out.rdata);
             --     write(l, string'("  s_data : "));
             --   hwrite(l, s_data(i));      
             --   writeline(output, l);          
             --end if;
            assert (s_axi_s2m_out.rdata = s_data(i) and s_axi_s2m_out.rid = s_id)
            report "rid of rdata mismatch !!!" & integer'image(i) 
            severity failure; 
          end loop;
    end loop; -- not eos
    wait;
  end process;

end architecture sim;
