--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the modules  axi_2vci and vci_2_axi
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- @date 2015-02-16

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library axi_lib;
use axi_lib.axi_pkg.all;

library vci_ram_lib;
use vci_ram_lib.all;

entity axilite_vci_sim is

  generic(n_t: positive := 1000000;      -- Number of test vectors
          rnd_ce: boolean := false;
          debug: boolean := false);
end entity axilite_vci_sim;

architecture sim of axilite_vci_sim is
  signal clk, srstn, ce    : std_ulogic;
  signal eos               : boolean;

  signal s_vci_i2t         : vci_i2t;
  signal s_vci_t2i         : vci_t2i;
  signal s_axilite_m2s_in  : axilite_gp_m2s;
  signal s_axilite_s2m_out : axilite_gp_s2m;
  signal s_axi_m2s_out     : axi_gp_m2s; 
  signal s_axi_s2m_in      : axi_gp_s2m;

  signal s_addr            : addr_t;
  signal s_data            : data_t;
  signal s_be              : be_t;

begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l        : line;
 
  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

   for i in 1 to n_t  loop
     if (i * 10) mod n_t = 0 then
         write(l, i);
         write(l, string'("/"));
         write(l, n_t);
         writeline(output, l);
     end if;
     wait until rising_edge(clk) and ce='1'; 

   end loop;

   eos <= true;
   write(l, string'("*** regression test passed"));
   writeline(output, l);
   wait;

  end process;

ivci_2_axi : entity work.vci_2_axi4(rtl) 
port map(
       clk         => clk           , 
       srstn       => srstn         , 
       ce          => ce            , 

       -- Axi4 out 
       p_axi_in    => s_axi_s2m_in  , -- in   
       p_axi_out   => s_axi_m2s_out , -- out  

       -- Vci in  
       p_vci_out   => s_vci_t2i     , -- out 
       p_vci_in    => s_vci_i2t       -- in  

);

iaxi4_2_vci : entity work.axi4lite_2_vci(rtl)
port map(
       clk         => clk           , 
       srstn       => srstn         , 
       ce          => ce            ,

       -- Axi4 in
       p_axi_in    => s_axilite_m2s_in  , -- in   
       p_axi_out   => s_axilite_s2m_out , -- out  

       -- Vci out 
       p_vci_out   => s_vci_i2t     , -- out 
       p_vci_in    => s_vci_t2i       -- in  
);


iaxi_ram: entity vci_ram_lib.axi_ram(rtl)
  port map(
           clk        => clk,
           srstn      => srstn,
           ce         => ce,

           p_axi_in   => s_axi_m2s_out ,
           p_axi_out  => s_axi_s2m_in  
);
 
  stim : process
  variable tmps : natural;
  begin

    
    while not eos loop
            s_axilite_m2s_in <= axilite_gp_m2s_none; 
 
            tmps := integer_rnd(10,50);
            for i in 1 to tmps loop 
                  wait until rising_edge(clk) and ce = '1';
            end loop;          

            s_addr              <= (others => '0');
            s_addr(15 downto 2) <= std_ulogic_vector_rnd(14);

            s_data <= std_ulogic_vector_rnd(DATA_SIZE);
            s_be   <= (others => '1');

            tmps := integer_rnd(10,50);
            for i in 1 to tmps loop 
                  wait until rising_edge(clk) and ce = '1';
            end loop;          

            -- send AXI ADDR WRITE REQ 
            s_axilite_m2s_in.awaddr  <= s_addr; 
            s_axilite_m2s_in.awvalid <= '1';

            wait until rising_edge(clk) and ce = '1' and s_axilite_s2m_out.awready = '1';
            s_axilite_m2s_in <= axilite_gp_m2s_none;

            -- send AXI DATA WRITE BURST
            s_axilite_m2s_in.wdata  <= s_data;
            s_axilite_m2s_in.wstrb  <= s_be; 
            s_axilite_m2s_in.wvalid <= '1';
            wait until rising_edge(clk) and ce = '1' and s_axilite_s2m_out.wready = '1';

            s_axilite_m2s_in        <= axilite_gp_m2s_none; 
            s_axilite_m2s_in.bready <= '1';   -- rspack

            wait until rising_edge(clk) and ce = '1' and s_axilite_s2m_out.bvalid = '1';
            s_axilite_m2s_in.bready <= '0';   -- rspack

            tmps := integer_rnd(10,50);
            for i in 1 to tmps loop 
                  wait until rising_edge(clk) and ce = '1';
            end loop; 

            -- send AXI READ REQ
            s_axilite_m2s_in.araddr    <= s_addr; 
            s_axilite_m2s_in.arvalid   <= '1';
             
            wait until rising_edge(clk) and ce = '1' and s_axilite_s2m_out.arready = '1';
            s_axilite_m2s_in        <= axilite_gp_m2s_none; 
            s_axilite_m2s_in.rready <= '1';   -- rspack

            -- Check Matching data
            wait until rising_edge(clk) and ce = '1' and s_axilite_s2m_out.rvalid = '1';
            assert s_axilite_s2m_out.rdata = s_data
            report "data mismatch !!!"  
            severity failure; 
    end loop; -- not eos
    wait;
  end process;

end architecture sim;
