--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the block cipher bc.vhd
-- @author Michele Di Giorgio (digiorgi@eurecom.fr)
-- @date 2014-03-11

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library des_lib;
use des_lib.des_pkg.all;

use work.bc_sim_pkg.all;

entity bc_sim is

  generic(n: positive := 10000;           -- Number of test vectors
          debug: boolean := false;
          rnd_ce: boolean := true);
  port(busy: out std_ulogic_vector(0 to 4);
       do: out des_word_64_array(0 to 4));

end entity bc_sim;

architecture sim of bc_sim is

  signal clk, ce, srstn, first, dsi: std_ulogic;
  signal busy_local: std_ulogic_vector(0 to 4);
  signal di, iv: des_word_64;
  signal do_local: des_word_64_array(0 to 4);
  signal key: desx_key;
  signal eos: boolean := false;
  signal mode: encryption_mode;

begin

  busy <= busy_local;
  do <= do_local;

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process  -- generate input stimuli
    variable l: line;
  begin
    srstn <= '0';
    di <= (others => '0');
    key.k1 <= (others => '0');
    key.k <= (others => '0');
    key.k2 <= (others => '0');
    mode <= encrypt;
    first <= '0';
    dsi <= '0';
    iv <= (others => '0');
    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce = '1';
    end loop;
    srstn <= '1';
    for i in 1 to n loop -- N tests
      if (i * 1000) mod n = 0 then -- Print advancement indicator
        write(l, i);
        write(l, string'("/"));
        write(l, n);
        writeline(output, l);
      end if;
      -- Randomly throw all inputs
      if i = 1 then
        dsi     <= '1';
        first   <= '1';
      else
        dsi     <= std_ulogic_rnd;
        first   <= std_ulogic_rnd;
      end if;
      if boolean_rnd then
        mode <= encrypt;
      else
        mode <= decrypt;
      end if;
      key.k1    <= std_ulogic_vector_rnd(64);
      key.k     <= std_ulogic_vector_rnd(56);
      key.k2    <= std_ulogic_vector_rnd(64);
      iv        <= std_ulogic_vector_rnd(64);
      di        <= std_ulogic_vector_rnd(64);
      wait until rising_edge(clk) and ce = '1';
    end loop;
    eos <= true;
    write(l, string'("Non-regression test passed"));
    writeline(output, l);
    wait;
  end process;

  g: for l2n in 0 to 4 generate

    ibc: entity work.bc
    generic map(l2n => l2n)
    port map(clk    => clk,
             ce     => ce,
             srstn  => srstn,
             first  => first,
             dsi    => dsi,
             iv     => iv,
             di     => di,
             key    => key,
             mode   => mode,
             busy   => busy_local(l2n),
             do     => do_local(l2n));

    process(clk, eos, srstn, ce, busy_local, dsi)
      variable l: line;
      variable cnt, min_cnt: natural;
      variable i: natural range 0 to 2;
    begin
      if rising_edge(clk) then
        if srstn = '0' then
          cnt := 0;
          min_cnt := 0;
          i := 0;
        elsif ce = '1' then
          if busy_local(l2n) = '0' and dsi = '1' then -- Start new DESX computation
            if i = 0 then
              i := 1;
            elsif i = 1 then
              min_cnt := cnt;
              i := 2;
            elsif cnt < min_cnt then
              min_cnt := cnt;
            end if;
            cnt := 0;
          else
            cnt := cnt + 1;
          end if;
        end if;
      end if;
      if eos then
        write(l, string'("l2n="));
        write(l, l2n);
        write(l, string'(": minimum II = "));
        write(l, min_cnt);
        writeline(output, l);
      end if;
    end process;


    process(clk) -- checking correctness
    variable tmp, di_prev: des_word_64;
    variable started: boolean;
    variable i: natural;
    begin
      if rising_edge(clk) then
        if srstn = '0' then
          started   := false;
          i         := 0;
          --
        elsif ce = '1' then
          if busy_local(l2n) = '0' then
            --print_desx(i, mode, first, key, di, iv, tmp, do_local(l2n));
            if started and do_local(l2n) /= tmp then
              print_desx(i, mode, first, key, di, iv, tmp, do_local(l2n));
              report "l2n=" & integer'image(l2n) & ": wrong output value" severity failure;
            end if;
            if dsi = '1' then
              i := i + 1;
              if mode = encrypt and first = '1' then
                tmp := desx(key, di xor iv, mode, debug);
              elsif mode = encrypt then
                tmp := desx(key, di xor tmp, mode, debug);
              elsif mode = decrypt and first = '1' then
                tmp := iv xor desx(key, di, mode, debug);
              else
                tmp := di_prev xor desx(key, di, mode, debug);
              end if;
              started := true;
              di_prev := di;    
            end if;
          end if;
       end if;
      end if;
    end process;

  end generate g;

end architecture sim;
