--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the pspe_cache module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- @date 2014-11-20

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

entity pspe_cache_sim is

generic(
      n_t         : natural := 100000;      -- Number of test vectors
      rnd_ce      : boolean := false;
      SETS        : natural := 4;
      BLOCKS      : natural := 2;
      debug       : boolean := false
);

end entity pspe_cache_sim;

architecture sim of pspe_cache_sim is

  signal clk         : std_ulogic;
  signal srstn       : std_ulogic;
  signal ce          : std_ulogic;
  signal eos         : boolean;
  signal s_cmd       : std_ulogic_vector(2 downto 0); -- 001 : read; 010 : write dir; 001 : write_data (write dir & data not exclusive)
  signal s_way_in    : std_ulogic_vector(SETS - 1 downto 0); -- one hot
  signal s_addr_in   : addr_t; 
  signal s_pspe_in   : pspe_t;
  signal s_data_in   : data_vector(BLOCKS - 1 downto 0);

  signal s_hit       : std_ulogic;
  signal s_data_out  : pspe_t;
  signal s_data_cmp  : data_vector(BLOCKS - 1 downto 0);
  signal s_way_out   : std_ulogic_vector(SETS - 1 downto 0);  

begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
  variable l         : line   ;

  begin
    eos   <= false;
    srstn <= '0';


    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

    for i in 1 to n_t loop
      if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
      end if;
      
      wait until rising_edge(clk) and ce = '1';

      if s_hit = '1' then
         for i in 0 to BLOCKS - 1 loop
            assert s_data_out(i) = s_data_cmp(i)
            report " hit but data_mismatch !!" severity failure;
         end loop;

       end if;
 
       assert isOneHot(unsigned(s_way_out)) 
       report "Way out is not OneHot" severity failure;
           
    end loop;

    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

gen_pspe_in : for i in s_pspe_in'range generate
      s_pspe_in(i) <= s_data_in(i);
end generate gen_pspe_in;

ipspe_cache : entity work.pspe_cache(rtl) 
generic map(
        SETS            => SETS,
        BLOCKS          => BLOCKS
) 
port map(
        clk             => clk        , 
        srstn           => srstn      , 
        ce              => ce         , 
                                
        cmd             => s_cmd      , 
                                
        set_in          => s_way_in   , 
        addr_in         => s_addr_in  , 
        data_in         => s_pspe_in  , 
                               
        hit             => s_hit      , 
        data_out        => s_data_out , 
        set_out         => s_way_out   
);

irnd_cache_gen : entity work.rnd_cache_gen(rtl) 
generic map(
        MEMSIZE         => 8192,
        WAYS            => SETS,
        BLOCKS          => BLOCKS
) 
port map(
        clk             => clk        , 
        srstn           => srstn      , 
        ce              => ce         , 
                                
        cmd             => s_cmd      , 
                                
        dirty_in        => open       , 
        way_in          => s_way_in   , 
        addr_in         => s_addr_in  , 
        data_in         => s_data_in  , 

        data_cmp        => s_data_cmp ,                               
        hit             => s_hit      , 
        way_out         => s_way_out  
);

end architecture sim;
