--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- rnd_cache_gen 
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-11-21

Library std;
use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library random_lib;
use random_lib.rnd.all;

entity rnd_cache_gen is
generic(
        MEMSIZE         : natural := 64;
        WAYS            : natural := 4;
        BLOCKS          : natural := 8   
);
port (
        clk             : in  std_ulogic;
        srstn           : in  std_ulogic;
        ce              : in  std_ulogic;
        
        cmd             : out  std_ulogic_vector(2 downto 0);                                           
 
        dirty_in        : out  std_ulogic;
        way_in          : out  std_ulogic_vector(WAYS - 1 downto 0);
        addr_in         : out  addr_t;                                                 
        data_in         : out  data_vector(BLOCKS - 1 downto 0);     
 
        data_cmp        : out  data_vector(BLOCKS - 1 downto 0);      
        
        way_out         : in   std_ulogic_vector(WAYS - 1 downto 0); 
        hit             : in   std_ulogic
);
end rnd_cache_gen;

architecture rtl of rnd_cache_gen is
constant BLOCK_BITS   : natural := log2_up(BLOCKS*CELL_SIZE);

type fsm   is (Init, Read, RdHit, Write);
type t_mem is array (MEMSIZE - 1 downto 0) of  data_t;

signal r_st  , s_st   : fsm;
signal s_mem , r_mem  : t_mem;
signal s_addr, r_addr : unsigned(ADDR_SIZE - 3 downto 0);
signal s_hit , r_hit  : std_ulogic;
signal s_way , r_way  : std_ulogic_vector(WAYS - 1 downto 0);
signal s_data, r_data : data_vector(BLOCKS - 1 downto 0);

-------------------
begin  

pTrans : process(r_st, hit, way_out, r_hit, r_way, r_mem, r_addr, r_data)

begin

s_st   <= r_st;
s_hit  <= r_hit;
s_way  <= r_way;
s_mem  <= r_mem;
s_addr <= r_addr;
s_data <= r_data;
 
case r_st is
   when Init =>
         for i in s_mem'range loop
            s_mem(i) <= std_ulogic_vector_rnd(DATA_SIZE);
         end loop;

         s_st <= Read;

    when Read =>
         s_st <= RdHit;

    when RdHit =>

        s_way <= way_out;

        for i in 0 to BLOCKS - 1 loop
           s_data(i) <= std_ulogic_vector_rnd(DATA_SIZE);
        end loop;
       
        s_hit   <= hit;
        s_st    <= Write;

   when Write =>
        for i in 0 to BLOCKS - 1 loop
           s_mem(to_integer(r_addr+i)) <= r_data(i);
        end loop;

        s_st    <= Read;
        s_addr  <= to_unsigned(integer_rnd(0, MEMSIZE - 1), s_addr'length)(s_addr'left downto log2_up(BLOCKS)) & (log2_up(BLOCKS) - 1 downto 0 => '0'); 
          
end case;
    
end process;

pGen : process(r_st, r_addr, r_hit, r_way, r_mem, r_data)
begin

addr_in  <= (others => '0');
cmd      <= (others => '0');
data_in  <= (others => (others => '0'));
data_cmp <= (others => (others => '0'));
way_in   <= (others => '0');
dirty_in <= '0';

case r_st is 
     when Init =>
          null;

     when Read =>
          addr_in  <= std_ulogic_vector(r_addr) & "00";
          cmd(2)   <= '1';

     when RdHit =>
          for i in 0 to BLOCKS - 1 loop
             data_cmp(i) <= r_mem(to_integer(r_addr+i));
          end loop;

     when Write =>
          data_in   <= r_data;

          addr_in   <= std_ulogic_vector(r_addr) & "00"; 
          cmd(1)    <= not r_hit;
          cmd(0)    <= '1';  
          dirty_in  <= '1';  

          way_in    <= r_way;
end case;
end process;
-------
pclk : process(clk)
begin

if rising_edge(clk) then
    if(srstn = '0') then
        r_st    <= Init;                          
	r_mem   <= (others => (others => '0'));
	r_data  <= (others => (others => '0'));
	r_addr  <= (others => '0');
	r_way   <= (others => '0');
	r_hit   <= '0';

     elsif ce = '1' then
	r_st    <= s_st  ; 
        r_mem   <= s_mem ;
        r_data  <= s_data;
	r_addr  <= s_addr;
	r_way   <= s_way ;
	r_hit   <= s_hit ;
     end if;
end if;

end process;
-----

end rtl;

