--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the cryptoInt module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- @date 2015-01-26

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library des_lib;
use des_lib.des_pkg.all;

entity cryptoInt_sim is

generic(
      n_t         : natural := 100000;      -- Number of test vectors
      rnd_ce      : boolean := false;
      debug       : boolean := false
);

end entity cryptoInt_sim;

architecture sim of cryptoInt_sim is
  constant max_reg : natural := 2048;

  type data_mac is array(1 downto 0) of data_t;

  type reg_t is array(max_reg - 1 downto 0) of data_register_read;
  type mac_t is array(max_reg - 1 downto 0) of data_mac;
  type key_t is array(max_reg - 1 downto 0) of desx_key;   

  signal key              : key_t;
  signal mac              : mac_t;
  signal reg              : reg_t;

  signal clk              : std_ulogic;
  signal srstn            : std_ulogic;
  signal ce               : std_ulogic;
  signal eos              : boolean;
  signal s_rd_crypto_cmd  : crypto_cmd_out;  
  signal s_rd_crypto_data : data_vector(11 downto 0);   -- up to 6 MAC Sets 
 
  signal s_wr_crypto_cmd  : crypto_cmd_in;              -- cmdack, rspval
  signal s_wr_crypto_data : data_vector(1 downto 0);
 
  signal idx              : natural;
begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
  variable l         : line   ;

  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

    for i in 1 to n_t loop
      if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
      end if;
      
      wait until rising_edge(clk) and ce = '1';

    end loop;

    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

icryptoInt : entity work.cryptoInt(rtl) 
port map(
        clk             => clk        , 
        srstn           => srstn      , 
        ce              => ce         , 
                                
       rd_crypto_cmd    => s_rd_crypto_cmd ,   
       rd_crypto_data   => s_rd_crypto_data,   
                                            
       wr_crypto_cmd    => s_wr_crypto_cmd ,   
       wr_crypto_data   => s_wr_crypto_data   
);
verif : process
begin

   idx <= 0;
   for i in 0 to max_reg - 1 loop
       mac(i)      <= (others => (others => '0'));
       reg(i).data <= (others => (others => '0'));
       key(i)      <= key_null;
   end loop;

   s_rd_crypto_cmd.cmdval    <= '0';  
   s_rd_crypto_cmd.mode      <= CRY_MAC;
   s_rd_crypto_cmd.key       <= key_null;
   s_rd_crypto_cmd.iv        <= (others => '0');  -- not used
   s_rd_crypto_cmd.size      <= (others => '0');      
   s_rd_crypto_cmd.cipher    <= '0';
   s_rd_crypto_cmd.id_src    <= (others => '0');
   s_rd_crypto_cmd.id_dest   <= (others => '0');
   s_rd_crypto_cmd.address   <= (others => '0');

  for i in 1 to 50 loop
      wait until rising_edge(clk) and ce='1';
  end loop;

  for i in 0 to max_reg - 1 loop 
    reg(i).addr(ADDR_SIZE - 1 downto LINE_BITS) <= std_ulogic_vector_rnd(ADDR_SIZE - LINE_BITS);
    reg(i).addr(LINE_BITS - 1 downto 0)         <= (others => '0');

    for j in 0 to SIZE_LINE_WORDS - 1 loop
        reg(i).data(j) <= std_ulogic_vector_rnd(DATA_SIZE);
        reg(i).be(j)   <= (others => '0');                    
    end loop;
    key(i).k1    <= std_ulogic_vector_rnd(64);
    key(i).k2    <= std_ulogic_vector_rnd(64);
    key(i).k     <= std_ulogic_vector_rnd(56);

  end loop;

  wait until rising_edge(clk) and ce = '1';

  for i in 0 to max_reg  - 1 loop
  -- compute mac
            s_rd_crypto_cmd.cmdval    <= '1';  
            s_rd_crypto_cmd.mode      <= CRY_MAC;
            s_rd_crypto_cmd.key       <= key(i);
            s_rd_crypto_cmd.iv        <= (others => '0');
            s_rd_crypto_cmd.size      <= std_ulogic_vector(to_unsigned(SIZE_LINE_WORDS, 4));      
            s_rd_crypto_cmd.address   <= reg(i).addr;

            for j in 0 to SIZE_LINE_WORDS - 1 loop
                s_rd_crypto_data(j) <= reg(i).data(j);
            end loop;

            wait until rising_edge(clk) and ce = '1'; 
            s_rd_crypto_cmd.cmdval    <= '0';  

            wait until rising_edge(clk) and ce = '1' and s_wr_crypto_cmd.rspval = '1';
            for k in 0 to 1 loop
              mac(i)(k) <= s_wr_crypto_data(k);
            end loop;

            for k in 1 to 10 loop
                wait until rising_edge(clk) and ce='1';
            end loop;
   end loop; 

   -- verif mac
   while not eos loop

          idx <= integer_rnd(0, max_reg - 1);       
          wait until rising_edge(clk) and ce='1';
 
          s_rd_crypto_cmd.cmdval    <= '1';  
          s_rd_crypto_cmd.mode      <= CRY_MAC;
          s_rd_crypto_cmd.key       <= key(idx); 
          s_rd_crypto_cmd.iv        <= (others => '0');
          s_rd_crypto_cmd.size      <= std_ulogic_vector(to_unsigned(SIZE_LINE_WORDS, 4));      
          s_rd_crypto_cmd.address   <= reg(idx).addr;

          for j in 0 to SIZE_LINE_WORDS - 1 loop
              s_rd_crypto_data(j) <= reg(idx).data(j);
          end loop;

          wait until rising_edge(clk) and ce = '1'; 
          s_rd_crypto_cmd.cmdval    <= '0';  

          wait until rising_edge(clk) and ce = '1' and s_wr_crypto_cmd.rspval = '1';
          for k in 0 to 1 loop
              assert mac(idx)(k) = s_wr_crypto_data(k) 
              report "mac mismatch at idx : " & integer'image(idx) 
              severity failure; 
          end loop; 

          for k in 1 to 10 loop
              wait until rising_edge(clk) and ce='1';
          end loop;

   end loop;
end process;
 
end architecture sim;
