--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the io_input_ctrl module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-11-25

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.global.all;

library random_lib;
use random_lib.rnd.all;

library vci_lib;
use vci_lib.vci_pack.all;

library vci_input_lib;
use vci_input_lib.all;

library vci_io_target_lib;
use vci_io_target_lib.all;

entity io_input_ctrl_sim is

  generic(n_t: natural := 100000;      -- Number of test vectors
          n: natural := 4;             -- Number of VciInitiators
          max_rnd_cnt: natural := 5;   -- maximum vlaue for random counter
          rnd_ce: boolean := false;
          debug: boolean := false);
  port(
          p_vci_target_out: out vci_i2t
  );

end entity io_input_ctrl_sim;

architecture sim of io_input_ctrl_sim is

  signal clk, srstn, ce        : std_ulogic;
  signal eos                   : boolean;
  signal s_vci_ini_in          : vci_t2i;
  signal s_vci_ini_out         : vci_i2t;
  signal s_vci_tgt_in          : vci_t2i;
  signal s_vci_tgt_out         : vci_i2t;

  signal s_addr                : addr_t;
  signal s_mb_addr             : addr_t;
  signal s_size                : addr_t;

  signal s_rd_reg_0            : data_register_read;  
  signal s_sc_nbusy            : std_ulogic;  
  signal s_wr_reg_3            : data_register_write;
  signal s_wr_reg_0            : data_register_write;
  signal s_wr_sec_cmd          : security_rq_out;
  signal s_rd_sec_cmd          : security_rq_in;
  signal s_rd_ctx_register     : ctx_register_read;
  signal s_wr_ctx_cmd          : security_ctx_cmd_out;
  signal s_rd_ctx_cmd          : security_ctx_cmd_in;
  signal s_wr_io_handler       : io_block;
  signal s_io_handler_data_in  : data_out ;  -- 8 words (256 bits)
  signal s_mt_cmd              : integrity_ctrl_cmd;
  signal s_mt_ack              : integrity_ctrl_ack;
  signal s_io_ack              : std_ulogic; 
  signal s_busy                : std_ulogic; -- flag to status reg;
  signal s_rd_irq              : irq_out;
  signal s_wr_irq_ack          : std_ulogic;     

begin

  p_vci_target_out    <= s_vci_ini_out;
  s_mt_ack.cack       <= '1';
  s_mt_ack.uack       <= '1';
  s_mt_ack.iack       <= '1';
  s_mt_ack.alap_ack   <= '1';

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 2 ns;
    clk <= '1';
    wait for 2 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l: line;
  begin
    eos <= false;
    srstn <= '0';
    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;
    srstn <= '1';
    for i in 1 to n_t loop
    if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
    end if;

      wait until rising_edge(clk) and ce='1'; 
    end loop;
    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

  irnd_vci_target: entity vci_lib.rnd_vci_target(rtl)
  port map( 
            clk              => clk,
            srstn            => srstn,
            eos              => eos,
            ce               => ce,
            p_vci_target_req => s_vci_tgt_out,     
            p_vci_target_rsp => s_vci_tgt_in       

  );

  irnd_ctx_gen : entity vci_input_lib.rnd_ctx_gen
  generic map(
           MAX_CTX_LATENCY  => 2
  ) 
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,
           rd_ctx_register  => s_rd_ctx_register,
           wr_ctx_cmd       => s_wr_ctx_cmd     ,
           rd_ctx_cmd       => s_rd_ctx_cmd     
  );

  irnd_sec_gen : entity vci_input_lib.rnd_sec_gen 
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,
           sc_nbusy         => s_sc_nbusy   ,
           rd_reg_0         => s_rd_reg_0   ,
           wr_reg_0         => s_wr_reg_0   ,
           wr_reg_3         => s_wr_reg_3   ,
           rd_sec_cmd       => s_rd_sec_cmd ,
           wr_sec_cmd       => s_wr_sec_cmd ,
           p_vci_ini_out    => s_vci_tgt_out,
           p_vci_ini_in     => s_vci_tgt_in    
  );

  irnd_vci_initiator : entity vci_io_target_lib.rnd_vci_io_init(rtl)  -- random VCI Initiator
  port map( 
            clk                 => clk,
            srstn               => srstn,
            ce                  => ce,
            eos                 => eos,
            p_vci_initiator_rsp => s_vci_ini_in,
            p_vci_initiator_req => s_vci_ini_out
 );
  ivci_io_target : entity vci_io_target_lib.vci_io_target(rtl)
  port map(
       clk                     => clk,
       srstn                   => srstn,
       ce                      => ce,

       io_handler_data_in      => s_io_handler_data_in,
       io_ack                  => s_io_ack,            
                                                    
       busy                    => s_busy,              
                                                    
       rd_irq                  => s_rd_irq,            
       wr_irq_ack              => s_wr_irq_ack,        

       p_vci_io_tgt_in         => s_vci_ini_out,
       p_vci_io_tgt_out        => s_vci_ini_in,

       wr_io_handler           => s_wr_io_handler,

       wr_master_ctx_init      => open,

       p_enable                => open,
       p_b_address             => s_addr      ,
       p_mb_address            => s_mb_addr   ,
       p_size                  => s_size      ,
       p_cache_sp              => open,
       p_cache_pspe            => open,
       p_cache_ms              => open,
       p_cache_mt              => open,

       p_irq_c                 => open,
       p_irq_e                 => open
);

  iio_input_ctrl: entity work.io_input_ctrl(rtl)
  port map(
           clk                  => clk,
           srstn                => srstn,
           ce                   => ce,

           p_mb_address         => s_mb_addr             ,
           p_size               => s_size                ,
           p_b_address          => s_addr                ,
                                                       
           rd_ctx_register      => s_rd_ctx_register     ,
           rd_reg_0             => s_rd_reg_0            ,
           rd_io_handler        => s_wr_io_handler       ,
           sc_nbusy             => s_sc_nbusy            ,
           mt_cmd               => s_mt_cmd              ,
           mt_ack               => s_mt_ack              ,
           wr_reg_0             => s_wr_reg_0            ,
           wr_reg_3             => s_wr_reg_3            ,
           io_handler_data_out  => s_io_handler_data_in  ,
           io_ack               => s_io_ack              ,
           busy                 => s_busy                ,
           wr_ctx_cmd           => s_wr_ctx_cmd          ,
           rd_ctx_cmd           => s_rd_ctx_cmd          ,
           wr_sec_cmd           => s_wr_sec_cmd          ,
           rd_sec_cmd           => s_rd_sec_cmd          ,
           wr_irq               => s_rd_irq              ,
           rd_irq_ack           => s_wr_irq_ack          
);

end architecture sim;
