--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- rnd_mem_gen.vhd
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-11-24

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library random_lib;
use random_lib.rnd.all;

entity rnd_mem_gen is
port (
       clk                     : in  std_ulogic;
       srstn                   : in  std_ulogic;
       ce                      : in  std_ulogic;

       -- Data from Mt or Ms (ctrl/cache)
       rd_mac_data             : out data_vector(1 downto 0);     

       -- Data from reg bank
       rd_reg                  : out data_register_read_array(REG_MAX - 1 downto 0);   -- @ out[5] in spec  
      
       -- Data to reg Bank
       wr_reg                  : in  data_register_write_array(REG_MAX - 1 downto 0); -- @ in[5] in spec

       -- Request from Mem_Arbiter
       rd_mem_rq               : out memory_ctrl_cmd_out; 
       wr_mem_rq               : in  memory_ctrl_cmd_in  
 
);

end rnd_mem_gen;

architecture rtl of rnd_mem_gen is
type fsm_state is (IDLE, CMD, RSP);


signal s_state   , r_state    : fsm_state;
signal s_mem_rq  , r_mem_rq   : memory_ctrl_cmd_out;
signal s_reg_data, r_reg_data : data_vector(SIZE_LINE_WORDS - 1 downto 0); 
signal s_reg_be  , r_reg_be   : be_vector(SIZE_LINE_WORDS - 1 downto 0); 
signal s_mac_data, r_mac_data : data_vector(1 downto 0); 
------------------
begin


pClk : process(clk)
begin

if rising_edge(clk) then
        if srstn = '0' then
              r_state                <= IDLE;
              r_mem_rq.cmdval        <= '0';
              r_mem_rq.cmd           <= READ;
              r_mem_rq.start_address <= (others => '0');
              r_mem_rq.plen          <= (others => '0');
              r_mem_rq.id            <= (others => '0'); 
              r_mem_rq.direct_write  <= '0'; 
              r_reg_data             <= (others => (others => '0'));
              r_reg_be               <= (others => (others => '0'));
              r_mac_data             <= (others => (others => '0'));

        elsif ce = '1' then
              r_state        <= s_state;
              r_mem_rq       <= s_mem_rq;
              r_reg_data     <= s_reg_data;
              r_reg_be       <= s_reg_be  ;
              r_mac_data     <= s_mac_data; 
        end if;
end if;
end process;

pTransition : process(r_state, wr_mem_rq, r_mem_rq, r_mac_data, r_reg_data, r_reg_be)
variable v_wrd_f : integer;
variable v_addr  : addr_t;
variable v_plen  : integer;

begin

s_mem_rq   <= r_mem_rq;
s_state    <= r_state;
s_mac_data <= r_mac_data;
s_reg_data <= r_reg_data;
s_reg_be   <= r_reg_be;
   
case r_state is
       when IDLE =>
            v_wrd_f := integer_rnd(0, SIZE_LINE_WORDS - 1);
            v_plen  := integer_rnd(1, (SIZE_LINE_WORDS - v_wrd_f))*CELL_SIZE; 
            v_addr  := std_ulogic_vector_rnd(ADDR_SIZE);
            v_addr(LINE_BITS - 1 downto CELL_BITS) := std_ulogic_vector(to_unsigned(v_wrd_f, WORD_BITS));
            s_mem_rq.cmdval        <= std_ulogic_rnd;
            s_mem_rq.cmd           <= cmd_mem_e'val(integer_rnd(0,1));
            s_mem_rq.start_address <= v_addr;
            s_mem_rq.plen          <= std_ulogic_vector(to_unsigned(v_plen, PLEN_SIZE));
            s_mem_rq.id            <= std_ulogic_vector(to_unsigned(integer_rnd(1, 2**REG_MAX - 1), REG_MAX));
            s_mem_rq.direct_write  <= std_ulogic_rnd;

            for i in s_mac_data'range loop
                s_mac_data(i) <= std_ulogic_vector_rnd(DATA_SIZE);
            end loop;

            for i in s_reg_data'range loop
                s_reg_data(i) <= std_ulogic_vector_rnd(DATA_SIZE);
                s_reg_be(i)   <= std_ulogic_vector_rnd(CELL_SIZE);
            end loop;
            
            s_state                <= CMD;                 
 
       when CMD =>
            if r_mem_rq.cmdval = '0' then
                s_state <= IDLE;
            elsif wr_mem_rq.cmdack = '1' then
                s_state <= RSP;
            else
                s_state <= CMD;
            end if;                 
 
       when RSP =>
            if wr_mem_rq.rspval = '1' then
               s_state <= IDLE;
            else
               s_state <= RSP;
            end if;
end case;

end process;

pGeneration : process(r_state, r_mem_rq, r_reg_data, r_mac_data, r_reg_be)

begin

rd_mem_rq.cmdval        <= '0';
rd_mem_rq.cmd           <= READ;
rd_mem_rq.start_address <= (others => '0');
rd_mem_rq.plen          <= (others => '0');
rd_mem_rq.id            <= (others => '0'); 
rd_mem_rq.direct_write  <= '0'; 

rd_mac_data             <= (others => (others => '0'));



case r_state is
       when IDLE =>                  
           for i in wr_reg'range loop       
                 rd_reg(i).data <= (others => (others => '0'));
                 rd_reg(i).addr <= (others => '0');
                 rd_reg(i).be   <= (others => (others => '0'));
           end loop;                                           

       when CMD =>

            rd_mem_rq.cmdval        <= r_mem_rq.cmdval       ;
            rd_mem_rq.cmd           <= r_mem_rq.cmd          ;
            rd_mem_rq.start_address <= r_mem_rq.start_address;
            rd_mem_rq.id            <= r_mem_rq.id           ;
            rd_mem_rq.direct_write  <= r_mem_rq.direct_write ;

            if r_mem_rq.cmd = WRITE and r_mem_rq.direct_write = '1' then
                  rd_mac_data    <= r_mac_data;
                  rd_mem_rq.plen <= std_ulogic_vector(to_unsigned(2*CELL_SIZE, PLEN_SIZE));
            else
                  rd_mem_rq.plen                             <= r_mem_rq.plen;
                  rd_reg(ctz(unsigned(r_mem_rq.id))).addr    <= line_aligned(r_mem_rq.start_address);
                  rd_reg(ctz(unsigned(r_mem_rq.id))).data    <= r_reg_data;
                  rd_reg(ctz(unsigned(r_mem_rq.id))).be      <= r_reg_be;
            end if;

       when RSP =>
            null;
end case;

end process;

end rtl;


