--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the ms_ctrl module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-12-07 

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;

library global_lib;
use global_lib.global.all;
use global_lib.numeric_std.all;

library crypto_lib;
use crypto_lib.all;

library mem_ctrl_lib;
use mem_ctrl_lib.all;

library vci_ram_lib;
use vci_ram_lib.all;

library random_lib;
use random_lib.rnd.all;

entity ms_ctrl_sim is

  generic(n_t    : natural := 100000;    -- Number of test vectors
          rnd_ce : boolean := false;
          debug  : boolean := false);
--port(
--);

end entity ms_ctrl_sim;

architecture sim of ms_ctrl_sim is
  signal clk, srstn, ce        : std_ulogic;
  signal eos, eof              : boolean;

  signal s_p_cache             : std_ulogic;
  signal s_rd_ms_cmd           : integrity_ctrl_cmd;
  signal s_rd_crypto_cmd       : crypto_cmd_in;       -- cmdack, rspval
  signal s_rd_crypto_data      : data_vector(1 downto 0);    
  signal s_rd_mem_rq           : memory_ctrl_cmd_in;  -- cmdack, rspval
  signal s_rd_reg_4            : data_register_read;  
  signal s_rd_irq_ack          : std_ulogic;      
  signal s_rd_reg_0            : data_register_read;  
  signal s_rd_reg_1            : data_register_read;  
  signal s_rd_reg_2            : data_register_read;  
  signal s_rd_ctx_register     : ctx_register_read;
  signal s_wr_ms_ack           : integrity_ctrl_ack;
  signal s_wr_crypto_cmd       : crypto_cmd_out;  
  signal s_wr_crypto_data      : data_vector(11 downto 0);       -- 6 MAC Sets 
  signal s_wr_mem_rq           : memory_ctrl_cmd_out; --  
  signal s_wr_mem_data         : direct_data;                      
  signal s_rd_mem_data         : data_vector(1 downto 0);          
  signal s_wr_irq              : irq_out;

  signal s_vci_ini_in          : vci_t2i;
  signal s_vci_ini_out         : vci_i2t;

  signal s_rd_reg              : data_register_read_array(REG_MAX - 1 downto 0);    
  signal s_wr_reg              : data_register_write_array(REG_MAX - 1 downto 0); 

  signal c       : natural;               
  file   ms_in  : TEXT;      

begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l        : line;
 
  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

   c <= 0;
   while not eof loop
     if (c * 10) mod n_t = 0 then
         write(l, c);
         write(l, string'("/"));
         write(l, n_t);
         writeline(output, l);
     end if;
     c <= c+1;
     wait until rising_edge(clk) and ce='1'; 

   end loop;

   eos <= true;
   write(l, string'("*** regression test passed"));
   writeline(output, l);
   wait;

  end process;
----
  ims_ctrl: entity work.ms_ctrl(rtl)
  port map(
           clk                  => clk,
           srstn                => srstn,
           ce                   => ce,

           p_cache              => s_p_cache         ,
           rd_ms_cmd            => s_rd_ms_cmd       ,
           rd_crypto_cmd        => s_rd_crypto_cmd   ,
           rd_crypto_data       => s_rd_crypto_data  ,
           rd_mem_rq            => s_rd_mem_rq       ,
           rd_reg_4             => s_rd_reg(4)       ,
           rd_irq_ack           => s_rd_irq_ack      ,
           rd_reg_0             => s_rd_reg_0        ,
           rd_reg_1             => s_rd_reg_1        ,
           rd_reg_2             => s_rd_reg_2        ,
           rd_ctx_register      => s_rd_ctx_register ,
           wr_ms_ack            => s_wr_ms_ack       ,
           wr_crypto_cmd        => s_wr_crypto_cmd   ,
           wr_crypto_data       => s_wr_crypto_data  ,
           wr_mem_rq            => s_wr_mem_rq       ,
           wr_mem_data          => s_wr_mem_data     ,
           wr_irq               => s_wr_irq           
       );

s_rd_mem_data <= s_wr_mem_data.data;

ivci_mem_ctrl: entity mem_ctrl_lib.vci_mem_ctrl(rtl)
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,
           p_srcid          => (others => '0'),
           rd_mac_data      => s_rd_mem_data ,
           rd_reg           => s_rd_reg      ,
           wr_reg           => s_wr_reg      ,
           rd_mem_rq        => s_wr_mem_rq   ,
           wr_mem_rq        => s_rd_mem_rq   ,
           p_vci_ini_out    => s_vci_ini_out ,
           p_vci_ini_in     => s_vci_ini_in  
);

ivci_ram: entity vci_ram_lib.vci_ram(rtl)
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,
           p_vci_ini_out    => s_vci_ini_out ,
           p_vci_ini_in     => s_vci_ini_in  
);

icryptoInt : entity crypto_lib.cryptoInt(rtl) 
port map(
        clk             => clk        , 
        srstn           => srstn      , 
        ce              => ce         , 
                                
       rd_crypto_cmd    => s_wr_crypto_cmd ,   
       rd_crypto_data   => s_wr_crypto_data,   
                                            
       wr_crypto_cmd    => s_rd_crypto_cmd ,   
       wr_crypto_data   => s_rd_crypto_data   
);
--
  stim : process
  variable good                : boolean;   -- Status of the read operations
  variable tmps                : natural;
  variable line_in, line_out   : Line; 
  variable field               : string(3 downto 1);
  variable v_p_cache           : std_ulogic;
  variable v_rd_ms_cmd         : integrity_ctrl_cmd;
  variable v_cmd               : std_ulogic_vector(1 downto 0);
  variable v_rd_irq_ack        : std_ulogic;    
  variable v_rd_ctx_register   : ctx_register_read;
  variable v_rd_reg_0          : data_register_read;  
  variable v_rd_reg_1          : data_register_read;  
  variable v_rd_reg_2          : data_register_read;  
 
  begin
  
    eof <= false;
    FILE_OPEN(ms_in,"ms_in.txt",READ_MODE);

    s_p_cache         <= '0';
    s_rd_irq_ack      <= '0';
    s_rd_ctx_register <= ctx_none;
 
    for i in 1 to 50 loop
      wait until rising_edge(clk) and ce = '1';  
    end loop;

    s_p_cache    <= '1';
    s_rd_irq_ack <= '1';
  
  --------
    while not endfile(ms_in) loop
          readline(ms_in, line_in);
          read(line_in, field, good);
          assert good report "Text I:O read field error" severity ERROR;

          v_rd_ctx_register := ctx_none;

          case field is
               when "--*"  =>
                    tmps := integer_rnd(50,100);
                    for i in 1 to tmps loop 
                      wait until rising_edge(clk) and ce = '1';
                    end loop;
 
               when "--#"  =>
                    wait until rising_edge(clk) and ce = '1' and (s_wr_ms_ack.cack='1' or s_wr_ms_ack.uack='1'); 
                    v_rd_ms_cmd.enable := '0';

               when "msc"  =>
                    read(line_in, v_rd_ms_cmd.enable);
                   hread(line_in, v_rd_ms_cmd.addr);
                    read(line_in, v_cmd);
                    read(line_in, v_rd_ms_cmd.src);
  
               when "rg0"  =>
                    hread(line_in, v_rd_reg_0.addr);
                    for i in 0 to 7 loop
                        hread(line_in, v_rd_reg_0.be(i));
                        hread(line_in, v_rd_reg_0.data(i));
                    end loop; 
  
               when "rg1"  =>
                    hread(line_in, v_rd_reg_1.addr);
                    for i in 0 to 7 loop
                        hread(line_in, v_rd_reg_1.be(i));
                        hread(line_in, v_rd_reg_1.data(i));
                    end loop;
  
               when "rg2"  =>
                    hread(line_in, v_rd_reg_2.addr);
                    for i in 0 to 7 loop
                        hread(line_in, v_rd_reg_2.be(i));
                        hread(line_in, v_rd_reg_2.data(i));
                    end loop;
   
               when "psx"  =>
                    for i in 0 to 1 loop
                        hread(line_in, v_rd_ctx_register.m_pspe(i));
                    end loop;
               when others =>
                    report "-- Unknown field : " & field severity FAILURE;
          end case;
   
          s_rd_reg_0            <= v_rd_reg_0          ;
          s_rd_reg_1            <= v_rd_reg_1          ;
          s_rd_reg_2            <= v_rd_reg_2          ;
          s_rd_ctx_register     <= v_rd_ctx_register   ;
          s_rd_ms_cmd           <= v_rd_ms_cmd         ;
          s_rd_ms_cmd.cmd       <= cmd_integrity_e'val(to_integer(unsigned(v_cmd)));
            
    end loop;

    write(line_out, "End of File encountered at " & time'image(now));
    writeline(output, line_out);
 
    wait until rising_edge(clk) and ce = '1';
    eof <= true;

    wait;
  end process;
--
  verif : process
   variable line_out : Line; 

   begin
           wait until rising_edge(clk) and ce='1';
           if (s_wr_irq.irq='1') then
  
                  hwrite(line_out, s_wr_irq.addr);
                   write(line_out , string'("  "));
                   write(line_out , io_status_errt_e'pos(s_wr_irq.typ));
                   write(line_out , string'("  "));
                   write(line_out , io_status_errc_e'pos(s_wr_irq.op));
                   writeline(output,line_out);     
                   report "IRQ launched due to mac mismatch !!!" 
                   severity FAILURE;

           end if;
 end process;

 pReg : process(clk)
 begin
    if rising_edge(clk) then
       if srstn = '0' then

          s_rd_reg(0).addr   <= (others =>'0');
          s_rd_reg(0).data   <= (others => (others => '0'));
          s_rd_reg(0).be     <= (others => (others => '0'));

          s_rd_reg(1).addr   <= (others =>'0');
          s_rd_reg(1).data   <= (others => (others => '0'));
          s_rd_reg(1).be     <= (others => (others => '0'));

          s_rd_reg(2).addr   <= (others =>'0');
          s_rd_reg(2).data   <= (others => (others => '0'));
          s_rd_reg(2).be     <= (others => (others => '0'));

          s_rd_reg(3).addr   <= (others =>'0');
          s_rd_reg(3).data   <= (others => (others => '0'));
          s_rd_reg(3).be     <= (others => (others => '0'));

          s_rd_reg(4).addr   <= (others =>'0');
          s_rd_reg(4).data   <= (others => (others => '0'));
          s_rd_reg(4).be     <= (others => (others => '0'));

        elsif ce = '1' then

             if s_wr_reg(4).enable = '1' then
                 if s_wr_reg(4).flush = '1' then
                    s_rd_reg(4).data <= (others => (others => '0'));
                    s_rd_reg(4).be   <= (others => (others => '0'));
                    s_rd_reg(4).addr <= (others => '1');                     
                 else                                                 
                    for i in s_wr_reg(4).be'range loop
                      if s_wr_reg(4).be(i) = X"F" then                 
                        s_rd_reg(4).data(i) <= s_wr_reg(4).data(i);
                        s_rd_reg(4).be(i)   <= s_wr_reg(4).be(i);    
                      end if;
                    end loop;
                    s_rd_reg(4).addr <= s_wr_reg(4).addr;
                 end if;
             end if;
         end if;
     end if;
end process;
  
end sim;
