--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the mt_cache_ctrl module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-12-07 

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;

library global_lib;
use global_lib.global.all;
use global_lib.numeric_std.all;

library random_lib;
use random_lib.rnd.all;

entity mt_cache_ctrl_sim is

  generic(n_t: natural := 10000000;    -- Number of test vectors
          n: natural := 4;             -- Number of VciInitiators
          max_rnd_cnt: natural := 5;   -- maximum vlaue for random counter
          rnd_ce: boolean := false;
          debug: boolean := false);
--port(
--);

end entity mt_cache_ctrl_sim;

architecture sim of mt_cache_ctrl_sim is
  signal clk, srstn, ce  : std_ulogic;
  signal eos, eof        : boolean;

  signal s_rd_cache_cmd  : cache_ctrl_cmd; 
  signal s_rd_cache_data : data_vector(1 downto 0);
  signal s_wr_cache_data : data_vector(1 downto 0);
  signal s_rd_mem_rq     : memory_ctrl_cmd_in;  -- cmdack, rspval
  signal s_rd_reg_4      : data_register_read;          
  signal s_wr_mem_rq     : memory_ctrl_cmd_out; --  
  signal s_wr_mem_data   : direct_data; -- data(2)                      
  signal s_wr_cache_ack  : cache_ctrl_ack;
  signal s_cache         : std_ulogic;
  signal s_direct_mem    : std_ulogic;

  signal c               : natural;               
  file   mtc_in          : TEXT;      

begin

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l        : line;
 
  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

   c <= 0;
   while not eof loop
     if (c * 10) mod n_t = 0 then
         write(l, c);
         write(l, string'("/"));
         write(l, n_t);
         writeline(output, l);
     end if;
     c <= c+1;
     wait until rising_edge(clk) and ce='1'; 

   end loop;

   eos <= true;
   write(l, string'("*** regression test passed"));
   writeline(output, l);
   wait;

  end process;
----
  imt_cache_ctrl: entity work.mt_cache_ctrl(rtl)
  port map(
           clk                  => clk,
           srstn                => srstn,
           ce                   => ce,

           p_cache              => s_cache         , -- in  std_ulogic;
           p_direct_mem         => s_direct_mem    , -- in  std_ulogic;
           rd_cache_cmd         => s_rd_cache_cmd  , -- in cache_ctrl_cmd; 
           rd_cache_data        => s_rd_cache_data , -- in data_vector(1 downto 0);
           wr_cache_ack         => s_wr_cache_ack  , -- out cache_ctrl_ack;
           wr_cache_data        => s_wr_cache_data , -- out data_vector(1 downto 0);
           wr_mem_rq            => s_wr_mem_rq     , -- out memory_ctrl_cmd_out; --  
           rd_mem_rq            => s_rd_mem_rq     , -- in  memory_ctrl_cmd_in;  -- cmdack, rspval
           wr_mem_data          => s_wr_mem_data   , -- out direct_data; -- data(2)                      
           rd_reg_4             => s_rd_reg_4        -- in  data_register_read  

       );

--
  stim : process
  variable good                : boolean;   -- Status of the read operations
  variable tmps                : natural;
  variable line_in, line_out   : Line; 
  variable field               : string(3 downto 1);
  variable v_rd_cache_cmd      : cache_ctrl_cmd;
  variable v_cmd               : natural;
  variable v_rd_cache_data     : data_vector(1 downto 0); 
  variable v_rd_reg_4          : data_register_read;  
  variable v_rd_mem_rq         : memory_ctrl_cmd_in;  
 
  begin
    s_cache      <= '0';
    s_direct_mem <= '1';
 
    eof <= false;
    FILE_OPEN(mtc_in,"mtc_in.txt",READ_MODE);

  --------
    while not endfile(mtc_in) loop
          readline(mtc_in, line_in);
          read(line_in, field, good);
          assert good report "Text I:O read field error" severity ERROR;

          case field is
               when "--*"  =>
                    tmps := integer_rnd(50,100);
                    for i in 1 to tmps loop 
                      wait until rising_edge(clk) and ce = '1';
                    end loop;
 
               when "--#"  =>
                    wait until rising_edge(clk) and ce = '1'; 
                    v_rd_cache_cmd.enable := '0';

               when "mtc"  =>
                    read(line_in, v_rd_cache_cmd.enable);
                   hread(line_in, v_rd_cache_cmd.address);
                    read(line_in, v_cmd);
                   hread(line_in, v_rd_cache_data(0));
                   hread(line_in, v_rd_cache_data(1));

               when "ack"  =>
                    read(line_in, v_rd_mem_rq.cmdack);
                    read(line_in, v_rd_mem_rq.rspval);

               when "rg4"  =>
                    hread(line_in, v_rd_reg_4.addr);
                    for i in 0 to 7 loop
                        hread(line_in, v_rd_reg_4.data(i));
                    end loop; 
  
               when others =>
                    report "-- Unknown field : " & field severity FAILURE;
          end case;
   
          s_rd_reg_4            <= v_rd_reg_4;
          s_rd_cache_data       <= v_rd_cache_data;
          s_rd_mem_rq           <= v_rd_mem_rq;
          s_rd_cache_cmd        <= v_rd_cache_cmd;
          s_rd_cache_cmd.cmd    <= cmd_mt_cache_e'val(v_cmd);
            
    end loop;

    write(line_out, "End of File encountered at " & time'image(now));
    writeline(output, line_out);
 
    wait until rising_edge(clk) and ce = '1';
    eof <= true;

    wait;
  end process;
 
end sim;
