--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- rnd_vci_initiator                                           
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-11-2
-- This module includes 3 architectures :
--  Random vci transaction generation (Read/Write) : rtl_All
--  Random vci Write aligned generation            : rtl_Wa
--  Random vci write non aligned generation        : rtl_Wn

LIBRARY ieee;
USE ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library vci_lib;
use vci_lib.vci_pack.all;

library random_lib;
use random_lib.rnd.all;

entity rnd_vci_initiator is
port(
       clk                 : in  std_ulogic;
       srstn               : in  std_ulogic;
       ce                  : in  std_ulogic;
       eos                 : in  boolean;
       p_vci_initiator_rsp : in  vci_t2i;
       p_vci_initiator_req : out vci_i2t
);
end entity rnd_vci_initiator;

------------------------------------------------
-- gen write burst vci transactions non aligned
------------------------------------------------
architecture rtl_Wn of rnd_vci_initiator is

type fsm is (IDLE, GEN_WR, SEND_WR, WAIT_RSP);

signal s_state  , r_state  : fsm;
signal s_buff   , r_buff   : vci_i2t;
signal s_addr   , r_addr   : unsigned(ADDR_SIZE - 1 downto 0);
signal s_wrds   , r_wrds   : integer;                          
signal s_wrds_2 , r_wrds_2 : integer;                          
signal s_idx    , r_idx    : integer;                          

begin

pwriteregs : process(clk)
begin
if rising_edge(clk) then
    if (srstn ='0') then
    	r_state   <= IDLE;             
    	r_buff    <= vci_i2t_none;
    	r_addr    <= (others => '0');
    	r_wrds    <= 0;
    	r_wrds_2  <= 0;
    	r_idx     <= 0;
    elsif ce = '1' then
    	r_state   <= s_state;
    	r_buff    <= s_buff;
    	r_addr    <= s_addr;
    	r_wrds    <= s_wrds;
    	r_wrds_2  <= s_wrds_2;
    	r_idx     <= s_idx;
    end if;
end if;
end process;


pTrans : process(r_state, eos, r_buff, r_addr, p_vci_initiator_rsp, r_idx, r_wrds, r_wrds_2)
variable v_wrd_f  : integer;
variable v_addr   : unsigned(ADDR_SIZE - 1 downto 0);
variable v_wrds   : integer;

begin

s_state  <= r_state;
s_addr   <= r_addr;
s_buff   <= r_buff;
s_idx    <= r_idx;
s_wrds   <= r_wrds;
s_wrds_2 <= r_wrds_2;

case r_state is
     when IDLE =>
        if not eos then
           s_buff  <= vci_i2t_none; 
           s_state <= GEN_WR;
        else
           s_state <= IDLE;
        end if;
     when GEN_WR =>

        v_wrd_f := integer_rnd(0, SIZE_LINE_WORDS - 1);
        v_addr  := unsigned_rnd(ADDR_SIZE);
        v_addr(LINE_BITS - 1 downto CELL_BITS) := to_unsigned(v_wrd_f, WORD_BITS);
        v_addr(CELL_BITS - 1 downto 0)         := (others => '0');

        v_wrds        := integer_rnd(1, SIZE_LINE_WORDS - v_wrd_f);

        s_buff        <= vci_req_gen; 
        s_buff.cmdval <= '1';
        s_buff.be     <= (others => '1');
        s_buff.addr   <= std_ulogic_vector(v_addr);
        s_buff.cmd    <= VCI_WRITE;

        if v_wrds = 1 then
            s_buff.eop <= '1';
        else
            s_buff.eop <= '0';
        end if;
       
        s_wrds        <= v_wrds; 
        s_wrds_2      <= v_wrds + v_wrd_f;
        s_addr        <= v_addr;
        s_state       <= SEND_WR;
        s_idx         <= 1; --v_wrd_f;

     when SEND_WR =>
        if p_vci_initiator_rsp.cmdack = '1' then

             if r_buff.eop = '1' then
                 s_state <= WAIT_RSP;
             else

                 v_addr        := r_addr + 4;

                 s_buff        <= vci_req_gen;
                 s_buff.cmdval <= '1';
                 s_buff.be     <= (others => '1');
                 s_buff.addr   <= std_ulogic_vector(v_addr);
                 s_buff.cmd    <= VCI_WRITE;
                 s_addr        <= v_addr; --r_addr + 4;
                 s_idx         <= r_idx + 1;
                 s_buff.eop    <= '0';

                 if r_idx = r_wrds - 1 then
 
                   if r_wrds_2 < SIZE_LINE_WORDS then
                        s_wrds_2 <= SIZE_LINE_WORDS;

                        v_wrd_f := integer_rnd(0, SIZE_LINE_WORDS - r_wrds_2 - 1);
                        v_addr  := unsigned_rnd(ADDR_SIZE);
                        v_addr(LINE_BITS - 1 downto CELL_BITS) := to_unsigned(v_wrd_f, WORD_BITS);
                        v_addr(CELL_BITS - 1 downto 0)         := (others => '0');
                        v_wrds        := integer_rnd(1, SIZE_LINE_WORDS - v_wrd_f);

                        s_wrds        <= v_wrds; 
                        s_addr        <= v_addr;
                        s_idx         <= 1; --v_wrd_f

                        if v_wrds = 1 then 
                            s_buff.eop <= '1'; 
                        end if;
                    else
                        s_buff.eop <= '1'; 
                    end if;
                end if;  
                s_state  <= SEND_WR;
           end if;
        else
           s_state <= SEND_WR;
        end if;

     when WAIT_RSP =>
        if p_vci_initiator_rsp.rspval = '1' then
           s_state <= IDLE;
        else
           s_state <= WAIT_RSP;
        end if;

end case;
end process;

pGen : process(r_state, r_buff)
begin

p_vci_initiator_req <= vci_i2t_none;

case r_state is
     when IDLE =>
          null;

     when GEN_WR =>
          null;

     when SEND_WR =>
          p_vci_initiator_req <= r_buff;

     when WAIT_RSP =>
          p_vci_initiator_req.rspack <= '1';

end case;
end process;

end rtl_Wn;

--------------------------------------------
-- gen write burst aligned vci transactions
--------------------------------------------
architecture rtl_Wa of rnd_vci_initiator is

type fsm is (IDLE, GEN_WR, SEND_WR, WAIT_RSP);

signal s_state, r_state : fsm;
signal s_buff , r_buff  : vci_i2t;
signal s_addr , r_addr  : unsigned(ADDR_SIZE - 1 downto 0);
signal s_wrds , r_wrds  : integer;                          
signal s_idx  , r_idx   : integer;                          

begin

pwriteregs : process(clk)
begin
if rising_edge(clk) then
    if (srstn ='0') then
    	r_state <= IDLE;             
    	r_buff  <= vci_i2t_none;
    	r_addr  <= (others => '0');
    	r_wrds  <= 0;
    	r_idx   <= 0;
    elsif ce = '1' then
    	r_state <= s_state;
    	r_buff  <= s_buff;
    	r_addr  <= s_addr;
    	r_wrds  <= s_wrds;
    	r_idx   <= s_idx;
    end if;
end if;
end process;

pTrans : process(r_state, eos, r_buff, r_addr, p_vci_initiator_rsp, r_idx, r_wrds)
variable v_wrd_f : integer;
variable v_addr  : unsigned(ADDR_SIZE - 1 downto 0);
variable v_wrds  : integer;

begin

s_state <= r_state;
s_addr  <= r_addr;
s_buff  <= r_buff;
s_idx   <= r_idx;
s_wrds  <= r_wrds;

case r_state is
     when IDLE =>
        if not eos then
           s_buff  <= vci_i2t_none; 
           s_state <= GEN_WR;
        else
           s_state <= IDLE;
        end if;
     when GEN_WR =>

        v_wrd_f := integer_rnd(0, SIZE_LINE_WORDS - 1);
        v_addr  := unsigned_rnd(ADDR_SIZE);
        v_addr(LINE_BITS - 1 downto CELL_BITS) := to_unsigned(v_wrd_f, WORD_BITS);
        v_addr(CELL_BITS - 1 downto 0)         := (others => '0');

        v_wrds        := integer_rnd(1, SIZE_LINE_WORDS - v_wrd_f);

        s_buff        <= vci_req_gen; 
        s_buff.cmdval <= '1';
        s_buff.be     <= (others => '1');
        s_buff.addr   <= std_ulogic_vector(v_addr);
        s_buff.cmd    <= VCI_WRITE;

        if v_wrds = 1 then
            s_buff.eop <= '1';
        else
            s_buff.eop <= '0';
        end if;
       
        s_wrds        <= v_wrds; 
        s_addr        <= v_addr;
        s_state       <= SEND_WR;
        s_idx         <= 1; --v_wrd_f;

     when SEND_WR =>
        if p_vci_initiator_rsp.cmdack = '1' then
           --s_buff <= vci_i2t_none; 
           if r_buff.eop = '1' then
              s_state <= WAIT_RSP;
           else

             v_addr        := r_addr + 4;

             s_buff        <= vci_req_gen;
             s_buff.cmdval <= '1';
             s_buff.be     <= (others => '1');
             s_buff.addr   <= std_ulogic_vector(v_addr);
             s_buff.cmd    <= VCI_WRITE;

             if r_idx = r_wrds - 1 then
                 s_buff.eop <= '1';
             else
                 s_buff.eop <= '0';
             end if;

             s_addr        <= v_addr; --r_addr + 4;
             s_idx         <= r_idx + 1;
             s_state       <= SEND_WR;
           end if;
        else
           s_state <= SEND_WR;
        end if;

     when WAIT_RSP =>
        if p_vci_initiator_rsp.rspval = '1' then
           s_state <= IDLE;
        else
           s_state <= WAIT_RSP;
        end if;

end case;
end process;

pGen : process(r_state, r_buff)
begin

p_vci_initiator_req <= vci_i2t_none;

case r_state is
     when IDLE =>
          null;

     when GEN_WR =>
          null;

     when SEND_WR =>
          p_vci_initiator_req <= r_buff;

     when WAIT_RSP =>
          p_vci_initiator_req.rspack <= '1';

end case;
end process;

end rtl_Wa;

---------------------------------------------
-- Gen random (Read/Write) vci transactions
--------------------------------------------
architecture rtl_All of rnd_vci_initiator is


type state_type is (IDLE, GEN_RD_REQ, GEN_WR_REQ, SEND_RD_REQ, SEND_WR_REQ, WAIT_RSP);
signal s_state, r_state : state_type;

signal s_buff , r_buff : vci_i2t;
signal s_cmd  , r_cmd  : std_ulogic; 

begin

pwriteregs : process(clk)
begin
if rising_edge(clk) then
    if (srstn ='0') then
    	r_state <= IDLE;
    	r_buff  <= vci_i2t_none;
    	r_cmd   <= '0';
    elsif ce = '1' and not eos then
    	r_state <= s_state;
    	r_buff  <= s_buff;
    	r_cmd   <= s_cmd;
    end if;
end if;
end process;
	
pTransition : process(r_state, r_buff, r_cmd, p_vci_initiator_rsp)
variable v_wrd_f : integer;
variable v_addr  : addr_t;
variable v_plen  : integer;

begin

s_state <= r_state;
s_buff  <= r_buff;
s_cmd   <= r_cmd;

case r_state is
	when IDLE =>  
                if r_cmd = '0' then
		    s_state <= GEN_RD_REQ;
                else
                    s_state <= GEN_WR_REQ;
                end if;
             
                s_cmd <= std_ulogic_rnd;

	when GEN_RD_REQ =>
               v_wrd_f := integer_rnd(0, SIZE_LINE_WORDS - 1);
               v_plen  := integer_rnd(1, (SIZE_LINE_WORDS - v_wrd_f))*CELL_SIZE; 
               v_addr  := std_ulogic_vector_rnd(ADDR_SIZE);
               v_addr(LINE_BITS - 1 downto CELL_BITS) := std_ulogic_vector(to_unsigned(v_wrd_f, WORD_BITS));

               s_buff         <= vci_req_gen; 
               s_buff.cmdval  <= '1'; 
               s_buff.addr    <= v_addr; 
               s_buff.plen    <= std_ulogic_vector(to_unsigned(v_plen, PLEN_SIZE));
               s_buff.cmd     <= VCI_READ;
               s_buff.eop     <= '1';
               s_state        <= SEND_RD_REQ;

       	when GEN_WR_REQ => 
               s_buff        <= vci_req_gen; 
               s_buff.cmdval <= '1'; 
               s_buff.cmd    <= VCI_WRITE;
               s_state       <= SEND_WR_REQ;

        when SEND_RD_REQ =>
               if p_vci_initiator_rsp.cmdack = '1' then
                  s_state <= WAIT_RSP;
               else
                  s_state <= SEND_RD_REQ;
               end if;
  
        when SEND_WR_REQ =>
                if p_vci_initiator_rsp.cmdack = '1' then
                   if r_buff.eop = '1' then
                      s_state <= WAIT_RSP;
                   else
                      s_state <= GEN_WR_REQ;
                   end if;
                else
                   s_state <= SEND_WR_REQ;
                end if;

         when WAIT_RSP =>
                if p_vci_initiator_rsp.rspval = '1' and p_vci_initiator_rsp.reop = '1' then
                   s_state <= IDLE;
                else
                   s_state <= WAIT_RSP;
                end if;
end case;
end process;

pgenMoore : process(r_state, r_buff)
begin

p_vci_initiator_req <= vci_i2t_none;

case r_state is

	when IDLE  =>
                null;				
	when GEN_RD_REQ =>
	        null;

	when GEN_WR_REQ =>
	        null;

	when SEND_RD_REQ  =>
                p_vci_initiator_req <= r_buff;

	when SEND_WR_REQ  =>
                p_vci_initiator_req <= r_buff;

	when WAIT_RSP =>
                p_vci_initiator_req.rspack <= '1';				

	when others =>

end case;

end process;

end rtl_All;


