--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the vci_input_ctrl module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-11-25

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.global.all;

library random_lib;
use random_lib.rnd.all;

library vci_lib;
use vci_lib.vci_pack.all;

entity vci_input_ctrl_sim is

  generic(n_t: natural := 100000;      -- Number of test vectors
          n: natural := 4;             -- Number of VciInitiators
          max_rnd_cnt: natural := 5;   -- maximum vlaue for random counter
          rnd_ce: boolean := false;
          debug: boolean := false);
  port(
          p_vci_target_out: out vci_i2t
  );

end entity vci_input_ctrl_sim;

architecture sim of vci_input_ctrl_sim is

  signal clk, srstn, ce    : std_ulogic;
  signal eos               : boolean;
  signal s_vci_ini_in      : vci_t2i;
  signal s_vci_ini_out     : vci_i2t;
  signal s_vci_tgt_in      : vci_i2t;
  signal s_vci_tgt_out     : vci_t2i;

  signal s_addr            : addr_t;
  signal s_mb_addr         : addr_t;
  signal s_size            : addr_t;

  signal s_rd_reg_0        : data_register_read;  
  signal s_sc_nbusy        : std_ulogic;  
  signal s_wr_reg_0        : data_register_write;
  signal s_wr_reg_3        : data_register_write;
  signal s_wr_sec_cmd      : security_rq_out;
  signal s_rd_sec_cmd      : security_rq_in;
  signal s_rd_ctx_register : ctx_register_read;
  signal s_wr_ctx_cmd      : security_ctx_cmd_out;
  signal s_rd_ctx_cmd      : security_ctx_cmd_in;
  signal s_mt_ack          : integrity_ctrl_ack;

begin

  p_vci_target_out    <= s_vci_ini_out;
  s_mt_ack.cack       <= '1';
  s_mt_ack.uack       <= '1';
  s_mt_ack.iack       <= '1';
  s_mt_ack.alap_ack   <= '1';

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 2 ns;
    clk <= '1';
    wait for 2 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l: line;
  begin
    eos <= false;
    srstn <= '0';
    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;
    srstn <= '1';
    for i in 1 to n_t loop
    if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
    end if;
    s_addr    <= std_ulogic_vector_rnd(ADDR_SIZE);
    s_mb_addr <= std_ulogic_vector_rnd(ADDR_SIZE);
    s_size    <= std_ulogic_vector_rnd(ADDR_SIZE);

      wait until rising_edge(clk) and ce='1' and s_vci_tgt_out.rspval='1' and s_vci_tgt_in.rspack='1'; -- request has been served
    end loop;
    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

  irnd_vci_target: entity vci_lib.rnd_vci_target(rtl)
  port map( 
            clk              => clk,
            srstn            => srstn,
            eos              => eos,
            ce               => ce,
            p_vci_target_req => s_vci_ini_out,     
            p_vci_target_rsp => s_vci_ini_in       

  );

  irnd_ctx_gen : entity work.rnd_ctx_gen
  generic map(
           MAX_CTX_LATENCY  => 10
  ) 
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,
           rd_ctx_register  => s_rd_ctx_register,
           wr_ctx_cmd       => s_wr_ctx_cmd     ,
           rd_ctx_cmd       => s_rd_ctx_cmd     
  );

  irnd_sec_gen : entity work.rnd_sec_gen 
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,
           sc_nbusy         => s_sc_nbusy   ,
           rd_reg_0         => s_rd_reg_0   ,
           wr_reg_0         => s_wr_reg_0   ,
           wr_reg_3         => s_wr_reg_3   ,
           rd_sec_cmd       => s_rd_sec_cmd ,
           wr_sec_cmd       => s_wr_sec_cmd ,
           p_vci_ini_out    => s_vci_ini_out,
           p_vci_ini_in     => s_vci_ini_in    
  );

  ivci_input_ctrl: entity work.vci_input_ctrl(rtl)
  port map(
           clk              => clk,
           srstn            => srstn,
           ce               => ce,

           rd_ctx_register  => s_rd_ctx_register,
           mt_ack           => s_mt_ack         ,
           rd_reg_0         => s_rd_reg_0       ,
           sc_nbusy         => s_sc_nbusy       ,
           p_mb_address     => s_mb_addr        ,
           p_size           => s_size           ,
           p_b_address      => s_addr           ,
           wr_reg_3         => s_wr_reg_3       ,
           wr_ctx_cmd       => s_wr_ctx_cmd     ,
           rd_ctx_cmd       => s_rd_ctx_cmd     ,
           wr_sec_cmd       => s_wr_sec_cmd     ,
           rd_sec_cmd       => s_rd_sec_cmd     , 
           p_vci_ini_in     => s_vci_tgt_in     ,
           p_vci_ini_out    => s_vci_tgt_out  
 );

  irnd_vci_initiator: entity work.rnd_vci_initiator(rtl_All)  -- random VCI Initiator
  port map( clk => clk,
            srstn => srstn,
            ce => ce,
            eos => eos,
            p_vci_initiator_rsp => s_vci_tgt_out, 
            p_vci_initiator_req => s_vci_tgt_in
  );
end architecture sim;
