--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the VciMerge module vci_merge.vhd
-- @author Michele Di Giorgio (digiorgi@eurecom.fr)
-- @date 2014-05-19

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.global.all;

library random_lib;
use random_lib.rnd.all;

library vci_lib;
use vci_lib.vci_pack.all;

entity vci_merge_sim is

  generic(n_t: positive := 100000;      -- Number of test vectors
          n: positive := 2;             -- Number of VciInitiators
          rnd_ce: boolean := false;
          debug: boolean := false);
  port(p_vci_initiator_out : out vci_t2i_vector(0 to n-1);
       p_vci_target_out    : out vci_i2t);

end entity vci_merge_sim;

architecture sim of vci_merge_sim is

  signal clk, srstn, ce: std_ulogic;
  signal eos: boolean;
  signal p_vci_initiator_in: vci_i2t_vector(0 to n-1);
  signal p_vci_target_in: vci_t2i;
  signal p_vci_initiator_out_local: vci_t2i_vector(0 to n-1);
  signal p_vci_target_out_local: vci_i2t;

  signal selected: std_ulogic_vector(0 to n-1);
  signal is_sel: boolean;

begin

  p_vci_initiator_out <= p_vci_initiator_out_local;
  p_vci_target_out <= p_vci_target_out_local;

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable idx_sel: natural range 0 to n-1;
    variable l: line;
  begin
    eos <= false;
    srstn <= '0';
    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;
    srstn <= '1';
    for i in 1 to n_t loop
      if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
      end if;
      for j in 0 to n-1 loop
        if selected(j) = '1' then
          idx_sel := j;
        end if;
      end loop;
      wait until rising_edge(clk) and ce='1' and p_vci_target_in.rspval='1' and p_vci_target_out_local.rspack='1'; -- request has been served
    end loop;
    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

  irnd_vci_target: entity vci_lib.rnd_vci_target(rtl)
  port map( clk => clk,
            srstn => srstn,
            eos => eos,
            ce => ce,
            p_vci_target_req => p_vci_target_out_local,
            p_vci_target_rsp => p_vci_target_in);

  g: for i in 0 to n-1 generate
    irnd_vci_initiator: entity vci_lib.rnd_vci_initiator(rtl)  -- random VCI Initiator
    port map( clk => clk,
              srstn => srstn,
              eos => eos,
              ce => ce,
              p_vci_initiator_rsp => p_vci_initiator_out_local(i),
              p_vci_initiator_req => p_vci_initiator_in(i));
  end generate g;

  ivci_merge: entity work.vci_merge(rtl)
  port map(
           clk             => clk,
           srstn           => srstn,
           ce              => ce,
           p_vci_ini_0_in  => p_vci_initiator_in(0),
           p_vci_ini_1_in  => p_vci_initiator_in(1),
           p_vci_tgt_in    => p_vci_target_in,
           p_vci_ini_0_out => p_vci_initiator_out_local(0),
           p_vci_ini_1_out => p_vci_initiator_out_local(1),
           p_vci_tgt_out   => p_vci_target_out_local);

  priority_checker: process(clk)
    variable idx_sel: natural range 0 to n-1;
  begin
    if rising_edge(clk) then
      if srstn = '0' then
        selected <= (others => '0');
        is_sel <= false;
      elsif ce = '1' then
        if not is_sel then
          -- if no one has been selected, select one
          idx_sel := 0;
          selected <= (others => '0');
          for i in n-1 to 0 loop
            if p_vci_initiator_in(i).cmdval = '1' then
              idx_sel := i;
              is_sel <= true;
            end if;
          end loop;
          selected(idx_sel) <= '1';
        else
          -- check that priority has been given to the right VCI Initiator
          for i in 0 to n-1 loop
            if p_vci_initiator_out_local(i).rspval='1' then
              assert selected(i) = '1'
                report "priority has been given to the wrong initiator"
                severity failure;
              is_sel <= false;   -- allow a new selection
            selected(i) <= '0';
          end if;
          end loop;
        end if; -- if not is_sel
      end if; -- if srstn
    end if; -- if clk
  end process;

end architecture sim;
