--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- axi_ram.vhd
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2015-02-16

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library axi_lib;
use axi_lib.axi_pkg.all;

entity axi_ram is
generic(
       RAM_ADDR_SIZE           : natural := 10
);
port (
       clk                     : in  std_ulogic;
       srstn                   : in  std_ulogic;
       ce                      : in  std_ulogic;

       p_axi_in                : in  axi_gp_m2s;
       p_axi_out               : out axi_gp_s2m 
);

end axi_ram;

architecture rtl of axi_ram is
type fsm_state is (IDLE, RD_ACK, RSP_RD, WR_ACK, CMD_WR, RSP_WR);

signal s_state       , r_state        : fsm_state;
signal s_words       , r_words        : unsigned(AXI_L - 1 downto 0); 
signal s_cnt         , r_cnt          : unsigned(AXI_L - 1 downto 0); 
signal s_addr        , r_addr         : unsigned(RAM_ADDR_SIZE - 1 downto 0);
signal s_srcid       , r_srcid        : std_ulogic_vector(AXI_GP_I - 1 downto 0);
signal s_data_in     , s_data_out     : std_ulogic_vector(AXI_GP_D - 1 downto 0);
signal s_be_in                        : std_ulogic_vector(AXI_GP_M - 1 downto 0);
signal s_eop                          : std_ulogic;
signal s_wen                          : std_ulogic;
------------------
begin

s_eop <= '1' when r_words = r_cnt else '0';

iram : entity work.ram
generic map (
	ADDR_SIZE   => RAM_ADDR_SIZE 
) 
port map (
	clk       => clk,
	ce        => ce,
	addr      => std_ulogic_vector(r_addr), 
	wdata     => s_data_in, 
	be        => s_be_in, 
	we        => s_wen,   
	rdata     => s_data_out   
 );

pClk : process(clk)
begin

if rising_edge(clk) then
        if srstn = '0' then
              r_state        <= IDLE;
              r_words        <= (others => '0');
              r_cnt          <= (others => '0');
              r_addr         <= (others => '0');
              r_srcid        <= (others => '0');
 
        elsif ce = '1' then
              r_state        <= s_state;
              r_words        <= s_words;
              r_cnt          <= s_cnt;
              r_addr         <= s_addr;
              r_srcid        <= s_srcid;
        end if;
end if;
end process;

pTransition : process(r_state, p_axi_in, s_eop, r_words, r_cnt, r_addr, r_srcid)

begin

s_state        <= r_state;
s_words        <= r_words;
s_cnt          <= r_cnt;
s_srcid        <= r_srcid;
s_addr         <= r_addr;
s_wen          <= '0';
s_data_in      <= (others => '0');
s_be_in        <= (others => '0');

case r_state is
       when IDLE =>
            s_addr  <= (others => '0');
            s_cnt   <= (others => '0');
            s_words <= (others => '0');

            if p_axi_in.arvalid = '1'  then
                s_words  <= (others => '0');                                     
                s_addr   <= unsigned(p_axi_in.araddr(RAM_ADDR_SIZE + 1 downto 2)); 
                s_cnt    <= unsigned(p_axi_in.arlen) + 1; 
                s_srcid  <= p_axi_in.arid;
              --s_size   <= p_axi_in.arsize;

                s_state  <= RD_ACK; 

            elsif p_axi_in.awvalid = '1' then
                s_state <= WR_ACK;
                s_addr  <= unsigned(p_axi_in.awaddr(RAM_ADDR_SIZE + 1 downto 2)); 
                s_srcid <= p_axi_in.awid;
            end if;
                 
      when RD_ACK =>
          s_addr  <= r_addr  + 1;
          s_words <= r_words + 1;
          s_state <= RSP_RD;          

       when RSP_RD =>
            if s_eop = '1' then
                s_state <= IDLE;  
            elsif (p_axi_in.rready = '1')  then  -- rspack 
                s_addr  <= r_addr  + 1;
                s_words <= r_words + 1;
                s_state <= RSP_RD;
            end if;

       when WR_ACK =>
            s_state <= CMD_WR;

       when CMD_WR =>

            if (p_axi_in.wvalid = '1')  then 
                s_addr    <= r_addr; 
                s_data_in <= p_axi_in.wdata; 
                s_be_in   <= p_axi_in.wstrb; 
                s_wen     <= '1';

                if (p_axi_in.wlast = '1') then
                    s_state <= RSP_WR;
                else
                    s_addr  <= r_addr + 1; -- awburst = INCR ?
                    s_state <= CMD_WR;
                end if; 
            end if;

       when RSP_WR =>
            if (p_axi_in.bready = '1') then -- rspack 
                s_state <= IDLE;
            end if;

end case;

end process;

pGeneration : process(r_state, s_eop, s_data_out, r_srcid)

begin

-- s2m
p_axi_out <= ('0',                    -- arready 
              (others => '0'),        -- rid      
              (others => '0'),        -- rdata
              (others => '0'),        -- rresp 
              '0',                    -- rlast 
              '0',                    -- rvalid
              '0',                    -- awready
              '0',                    -- wready
              (others => '0'),        -- bid      
              '0',                    -- bvalid 
              (others => '0')         -- bresp
             );
 
case r_state is
       when IDLE =>                  
            null;

       when RD_ACK =>
            p_axi_out.arready <= '1';   -- cmdack
 
       when RSP_RD =>
            p_axi_out.rvalid <= '1';
            p_axi_out.rlast  <= s_eop; 
            p_axi_out.rresp  <= (others => '0');
            p_axi_out.rdata  <= s_data_out; 
            p_axi_out.rid    <= r_srcid;

       when WR_ACK =>
            p_axi_out.awready <= '1';   -- cmdack

       when CMD_WR =>
            p_axi_out.wready  <= '1';   -- cmdack                                       
                   
       when RSP_WR =>
            p_axi_out.bvalid <= '1';
            p_axi_out.bresp  <= (others => '0');
            p_axi_out.bid    <= r_srcid;
end case;

end process;

end rtl;


