--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- vci_ram.vhd
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2014-09-17

library ieee;
use ieee.std_logic_1164.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

entity vci_ram is
generic(
       RAM_ADDR_SIZE           : natural := 14
);
port (
       clk                     : in  std_ulogic;
       srstn                   : in  std_ulogic;
       ce                      : in  std_ulogic;

       p_vci_ini_out           : in  vci_i2t;
       p_vci_ini_in            : out vci_t2i
);

end vci_ram;

architecture rtl of vci_ram is
type fsm_state is (IDLE, CMD_RD, CMD_WR, RSP_WR);

signal s_state       , r_state        : fsm_state;
signal s_words       , r_words        : unsigned(PLEN_SIZE - CELL_BITS - 1 downto 0); 
signal s_cnt         , r_cnt          : unsigned(PLEN_SIZE - CELL_BITS - 1 downto 0); 
signal s_addr        , r_addr         : unsigned(RAM_ADDR_SIZE - 1 downto 0);
signal s_srcid       , r_srcid        : srcid_t;
signal s_trdid       , r_trdid        : trdid_t;
signal s_pktid       , r_pktid        : pktid_t;
signal s_data_in     , s_data_out     : data_t;
signal s_be_in                        : be_t;
signal s_eop                          : std_ulogic;
signal s_wen                          : std_ulogic;
------------------
begin

s_eop <= '1' when r_words = r_cnt - 1 else
                 '0';

iram : entity work.ram
generic map (
	ADDR_SIZE   => RAM_ADDR_SIZE 
) 
port map (
	clk       => clk,
	ce        => ce,
	addr      => std_ulogic_vector(s_addr), 
	wdata     => s_data_in, 
	be        => s_be_in, 
	we        => s_wen,   
	rdata     => s_data_out   
 );

pClk : process(clk)
begin

if rising_edge(clk) then
        if srstn = '0' then
              r_state        <= IDLE;
              r_words        <= (others => '0');
              r_cnt          <= (others => '0');
              r_addr         <= (others => '0');
              r_srcid        <= (others => '0');
              r_trdid        <= (others => '0');
              r_pktid        <= (others => '0');
 
        elsif ce = '1' then
              r_state        <= s_state;
              r_words        <= s_words;
              r_cnt          <= s_cnt;
              r_addr         <= s_addr;
              r_srcid        <= s_srcid;
              r_trdid        <= s_trdid;
              r_pktid        <= s_pktid;
        end if;
end if;
end process;

pTransition : process(r_state, p_vci_ini_out, s_eop, r_words, r_cnt, r_addr, r_srcid, r_trdid, r_pktid)

begin

s_state        <= r_state;
s_words        <= r_words;
s_cnt          <= r_cnt;
s_srcid        <= r_srcid;
s_trdid        <= r_trdid;
s_pktid        <= r_pktid;

s_addr         <= r_addr;
s_wen          <= '0';
s_data_in      <= (others => '0');
s_be_in        <= (others => '0');

case r_state is
       when IDLE =>
            if p_vci_ini_out.cmdval = '1' then
                s_words        <= (others => '0');                                     
                s_addr         <= unsigned(p_vci_ini_out.addr(RAM_ADDR_SIZE + 1 downto 2)); 
                s_cnt          <= unsigned(p_vci_ini_out.plen(p_vci_ini_out.plen'left downto CELL_BITS));
                s_srcid        <= p_vci_ini_out.srcid;
                s_trdid        <= p_vci_ini_out.trdid;
                s_pktid        <= p_vci_ini_out.pktid;

                if p_vci_ini_out.cmd = VCI_READ then 
                      s_state <= CMD_RD;
                elsif p_vci_ini_out.cmd = VCI_WRITE then
                      s_data_in <= p_vci_ini_out.wdata; 
                      s_be_in   <= p_vci_ini_out.be; 
                      s_wen     <= '1';

                      if p_vci_ini_out.eop = '1' then 
                          s_state <= RSP_WR;
                      else 
                          s_state <= CMD_WR;
                      end if;
                end if;
            else
                s_state <= IDLE;
            end if;
                  
       when CMD_RD =>
            if s_eop = '1' then
                s_state <= IDLE;  
            elsif (p_vci_ini_out.rspack = '1')  then 
                s_addr  <= r_addr  + 1;
                s_words <= r_words + 1;
                s_state <= CMD_RD;
            else 
                s_state <= CMD_RD;
            end if;

       when CMD_WR =>

            if (p_vci_ini_out.cmdval = '1')  then 
                s_addr    <= unsigned(p_vci_ini_out.addr(RAM_ADDR_SIZE + 1 downto 2)); 
                s_data_in <= p_vci_ini_out.wdata; 
                s_be_in   <= p_vci_ini_out.be; 
                s_wen     <= '1';

                if (p_vci_ini_out.eop = '1') then
                    s_state     <= RSP_WR;
                else
                    s_state     <= CMD_WR;
                end if; 
            else 
                s_state <= CMD_WR;
            end if;

       when RSP_WR =>
               if (p_vci_ini_out.rspack = '1') then 
                      s_state <= IDLE;
               else
                      s_state <= RSP_WR;
               end if;

end case;

end process;

pGeneration : process(r_state, s_eop, s_data_out, r_srcid, r_trdid, r_pktid)

begin

p_vci_ini_in     <= vci_t2i_none;


case r_state is
       when IDLE =>                  
            p_vci_ini_in.cmdack  <= '1';                                          

       when CMD_RD =>

            p_vci_ini_in.rspval <= '1';
            p_vci_ini_in.reop   <= s_eop; 
            p_vci_ini_in.rerror <= (others => '0');
            p_vci_ini_in.rdata  <= s_data_out; 
            p_vci_ini_in.rsrcid <= r_srcid;
            p_vci_ini_in.rtrdid <= r_trdid;
            p_vci_ini_in.rpktid <= r_pktid;

       when CMD_WR =>
            p_vci_ini_in.cmdack  <= '1';                                         
                   
       when RSP_WR =>
            p_vci_ini_in.rspval <= '1';
            p_vci_ini_in.reop   <= '1';
            p_vci_ini_in.rerror <= (others => '0');
            p_vci_ini_in.rdata  <= (others => '0');
            p_vci_ini_in.rsrcid <= r_srcid;
            p_vci_ini_in.rtrdid <= r_trdid;
            p_vci_ini_in.rpktid <= r_pktid;  
end case;

end process;

end rtl;


