--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the axi_secbus module
-- author Abdelmalek Si Merabet (abdelmalek.si-merabet@telecom-paristech.fr)
-- date : 2015-04-14 

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;

library global_lib;
use global_lib.global.all;
use global_lib.numeric_std.all;

library axi_lib;
use axi_lib.axi_pkg.all;

library random_lib;
use random_lib.rnd.all;

entity axi_secbus_sim is

  generic(n_t: natural := 10000000;    -- Number of test vectors
          n: natural := 4;             -- Number of VciInitiators
          max_rnd_cnt: natural := 5;   -- maximum vlaue for random counter
          rnd_ce: boolean := false;
          debug: boolean := false);
--port(
--);

end entity axi_secbus_sim;

architecture sim of axi_secbus_sim is
  constant axi_s2m_none : axi_gp_s2m := ('0', (others => '0'), (others => '0'), (others => '0'),        
                                         '0', '0', '0', '0', (others => '0'), '0', (others => '0'));

  constant axi_m2s_none  : axi_gp_m2s :=  
                            ((others => '0'), (others => '0'), (others => '0'), (others => '0'),  
                             (others => '0'), (others => '0'), (others => '0'), (others => '0'),  
                             (others => '0'), '0'            , '0'            , (others => '0'),    
                             (others => '0'), (others => '0'), (others => '0'), (others => '0'), 
                             (others => '0'), (others => '0'), (others => '0'), (others => '0'), 
                             '0'            , (others => '0'), (others => '0'), (others => '0'), 
                             '0'            , '0'            , '0'           );

  constant axilite_m2s_none : axilite_gp_m2s :=  
                            ((others => '0'), (others => '0'), '0', '0',  
                             (others => '0'), (others => '0'), '0',   
                             (others => '0'), (others => '0'), '0', '0');

  signal clk, srstn, ce        : std_ulogic;
  signal eos, eof              : boolean;
  signal eof_ini, eof_tgt      : boolean;

  signal s_axi_ini_in          : axi_gp_m2s;
  signal s_axi_ini_out         : axi_gp_s2m;
               
  signal s_axi_io_in           : axilite_gp_m2s;
  signal s_axi_io_out          : axilite_gp_s2m;
               
  signal s_axi_tgt_out         : axi_gp_m2s; 
  signal s_axi_tgt_in          : axi_gp_s2m;

  signal s_irq_c               : std_ulogic;
  signal s_irq_e               : std_ulogic;

  signal c       : natural;               
  file   axi_ini_in  : TEXT;      
  file   axi_tgt_in  : TEXT;      

begin

  eof <= eof_ini and eof_tgt;

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l        : line;
 
  begin
    eos   <= false;
    srstn <= '0';

    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce='1';
    end loop;

    srstn <= '1';

   c <= 0;
   while not eof loop
     if (c * 10) mod n_t = 0 then
         write(l, c);
         write(l, string'("/"));
         write(l, n_t);
         writeline(output, l);
     end if;
     c <= c+1;
     wait until rising_edge(clk) and ce='1'; 

   end loop;

   eos <= true;
   write(l, string'("*** regression test passed"));
   writeline(output, l);
   wait;

  end process;
----
iaxi_secbus_wrapper : entity work.axi_secbus_wrapper(netlist)
port map(
       clk           => clk,     
       srstn         => srstn,   
       ce            => ce,      

       p_axi_ini_in  => s_axi_ini_in,    -- in  axi_gp_m2s;
       p_axi_ini_out => s_axi_ini_out,   -- out axi_gp_s2m;
                                    
       p_axi_io_in   => s_axi_io_in,     -- in  axilite_gp_m2s;
       p_axi_io_out  => s_axi_io_out,    -- out axilite_gp_s2m;
                                    
       p_axi_tgt_out => s_axi_tgt_out,   -- out axi_gp_m2s; 
       p_axi_tgt_in  => s_axi_tgt_in,    -- in  axi_gp_s2m;
                                    
       p_irq_c       => s_irq_c,         -- out std_ulogic;
       p_irq_e       => s_irq_e          -- out std_ulogic

);

--
  stim_cmd : process
  variable good                : boolean;   -- Status of the read operations
  variable tmps                : natural;
  variable line_in, line_out   : Line; 
  variable field               : string(3 downto 1);
  variable v_axi_ini_in        : axi_gp_m2s;
  variable v_axi_io_in         : axilite_gp_m2s;
  variable v_ini_cmd           : natural; 
  variable v_io_cmd            : natural; 
  
  begin
  
    eof_ini <= false;
    FILE_OPEN(axi_ini_in,"axi_ini_in.txt",READ_MODE);
 
    v_axi_ini_in := axi_m2s_none;
    v_axi_io_in  := axilite_m2s_none;

    for i in 1 to 50 loop
      wait until rising_edge(clk) and ce = '1';  
    end loop;
  --------
    while not endfile(axi_ini_in) loop
          readline(axi_ini_in, line_in);
          read(line_in, field, good);
          assert good report "Text I:O read field error" severity ERROR;
          case field is
               when "--*"  =>
                    tmps := integer_rnd(100,500);
                    for i in 1 to tmps loop 
                      wait until rising_edge(clk) and ce = '1';
                    end loop;

                    s_axi_ini_in.rready <= '0';
                    s_axi_io_in.rready  <= '0';
                    s_axi_ini_in.bready <= '0';
                    s_axi_io_in.bready  <= '0';

               when "--#"  =>
                    wait until rising_edge(clk) and ce = '1' and ((s_axi_ini_in.arvalid = '1' and s_axi_ini_out.arready = '1') or
                                                                  (s_axi_ini_in.awvalid = '1' and s_axi_ini_out.awready = '1'));

                    s_axi_ini_in.arvalid <= '0';
                    s_axi_ini_in.awvalid <= '0';

                    if s_axi_ini_in.arvalid = '1' then
                             s_axi_ini_in.rready <= '1';
                             wait until rising_edge(clk) and ce = '1' and s_axi_ini_out.rvalid='1' and s_axi_ini_out.rlast='1';
                    elsif s_axi_ini_in.wlast = '1' then
                              s_axi_ini_in.bready <= '1'; 
                              wait until rising_edge(clk) and ce = '1' and s_axi_ini_out.bvalid='1';                                                                                                                                   end if;
                              
               when "--$"  =>
                    wait until rising_edge(clk) and ce = '1' and 
                              ((s_axi_io_in.arvalid = '1' and s_axi_io_out.arready = '1') or  
                               (s_axi_io_in.awvalid = '1' and s_axi_io_out.awready = '1'));

                    s_axi_io_in.arvalid <= '0';
                    s_axi_io_in.awvalid <= '0';
                    s_axi_io_in.rready  <= '1';
                    s_axi_io_in.bready  <= '1';

                    wait until rising_edge(clk) and ce = '1' and (s_axi_io_out.rvalid='1' or s_axi_io_out.bvalid='1'); 
                 
               when "iot"  =>
                    read(line_in , v_axi_io_in.arvalid);
                    hread(line_in, v_axi_io_in.araddr);
                    read(line_in , v_axi_io_in.awvalid);
                    hread(line_in, v_axi_io_in.awaddr);
                    read(line_in, v_axi_io_in.wvalid);
                    hread(line_in, v_axi_io_in.wdata);
                    read(line_in , v_axi_io_in.wstrb);

                    s_axi_io_in      <= v_axi_io_in;

               when "tgt"  =>
                    read(line_in , v_axi_ini_in.arvalid);
                    hread(line_in, v_axi_ini_in.araddr);
                    read(line_in, v_axi_ini_in.arlen);
                    read(line_in , v_axi_ini_in.arburst);
                    read(line_in , v_axi_ini_in.awvalid);
                    hread(line_in, v_axi_ini_in.awaddr);
                    read(line_in, v_axi_ini_in.awlen);
                    read(line_in , v_axi_ini_in.awburst);
                    read(line_in, v_axi_ini_in.wvalid);
                    hread(line_in, v_axi_ini_in.wdata);
                    read(line_in , v_axi_ini_in.wstrb);
                    read(line_in , v_axi_ini_in.wlast);

                    s_axi_ini_in      <= v_axi_ini_in;

               when others =>
                    report "-- Unknown field : " & field severity FAILURE;
          end case;
           
    end loop;

    write(line_out, "End of File VCI_INI_IN encountered at " & time'image(now));
    writeline(output, line_out);
 
    wait until rising_edge(clk) and ce = '1';
    eof_ini <= true;

    wait;
  end process;

  stim_rsp   : process
  variable good                : boolean;   -- Status of the read operations
  variable tmps                : natural;
  variable line_in, line_out   : Line; 
  variable field               : string(3 downto 1);
  variable v_axi_tgt_in        : axi_gp_s2m; 
 
  begin
  
    eof_tgt <= false;
    FILE_OPEN(axi_tgt_in,"axi_tgt_in.txt",READ_MODE);
    s_axi_tgt_in   <= axi_s2m_none;    

  --------
    while not endfile(axi_tgt_in) loop
          wait until rising_edge(clk) and ce='1' and (s_axi_tgt_out.arvalid = '1' or s_axi_tgt_out.awvalid = '1');
          if s_axi_tgt_out.arvalid = '1' then
              s_axi_tgt_in.arready <= '1';

              wait until rising_edge(clk) and ce='1';
              s_axi_tgt_in.arready <= '0';
              
              tmps := integer_rnd(50,100);
              for i in 1 to tmps loop 
                 wait until rising_edge(clk) and ce = '1';
              end loop;
              
              while s_axi_tgt_in.rlast = '0' loop 
                    v_axi_tgt_in := axi_s2m_none;

                    readline(axi_tgt_in, line_in);
                    read(line_in, field, good);
                    assert good report "Text I:O read field error" severity ERROR;
                    
                    read(line_in, v_axi_tgt_in.rvalid);
                    hread(line_in, v_axi_tgt_in.rdata);
                    read(line_in, v_axi_tgt_in.rlast);
                    
                    s_axi_tgt_in  <= v_axi_tgt_in;
                    
                    wait until rising_edge(clk) and ce='1' and s_axi_tgt_out.rready = '1';
              end loop;
              s_axi_tgt_in <= axi_s2m_none;

          else -- awvalid
              s_axi_tgt_in.awready <= '1';
              s_axi_tgt_in.wready  <= '1';

              if s_axi_tgt_out.wlast = '1' then
                     wait until rising_edge(clk) and ce='1';
                     s_axi_tgt_in.awready <= '0';
                     s_axi_tgt_in.wready  <= '0';
                     
                     tmps := integer_rnd(50,100);
                     for i in 1 to tmps loop 
                        wait until rising_edge(clk) and ce = '1';
                     end loop;
                     
                     v_axi_tgt_in := axi_s2m_none;

                     readline(axi_tgt_in, line_in);
                     read(line_in, field, good);
                     assert good report "Text I:O read field error" severity ERROR;
 
                     read(line_in, v_axi_tgt_in.rvalid);
                     hread(line_in, v_axi_tgt_in.rdata);
                     read(line_in, v_axi_tgt_in.rlast);
                     read(line_in, v_axi_tgt_in.bvalid);
                     
                     s_axi_tgt_in  <= v_axi_tgt_in;
                           
                     wait until rising_edge(clk) and ce='1' and s_axi_tgt_out.bready = '1';
                     s_axi_tgt_in <= axi_s2m_none;
              end if;
          end if; -- arvalid or awvalid


    end loop;

    write(line_out, "End of File VCI_TGT_IN encountered at " & time'image(now));
    writeline(output, line_out);
 
    wait until rising_edge(clk) and ce = '1';
    eof_tgt <= true;

    wait;
  end process;

 verif : process
 begin
    while not eos loop
          wait until rising_edge(clk) and ce='1';
         --if (s_irq_c='1') then 
         --   report "IRQ of type C launched !!!" 
         --   severity failure;

          if (s_irq_e='1') then
             report "IRQ of type E launched !!!" 
             severity failure;
          end if;
    end loop;
    wait;
 end process;
 
end sim;
