--
-- SecBus ( http://secbus.telecom-paristech.fr/ ) - This file is part of SecBus
-- Copyright (C) 2009-2015 - Telecom ParisTech
-- Contacts: contact-secbus@telecom-paristech.fr
--
-- This file must be used under the terms of the CeCILL.
-- This source file is licensed as described in the file COPYING, which
-- you should have received as part of this distribution. The terms
-- are also available at
-- http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
--

-- @brief simulation environment for the VciSplit module vci_split.vhd
-- @author Michele Di Giorgio (digiorgi@eurecom.fr)
-- @date 2014-06-05

use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;

library random_lib;
use random_lib.rnd.all;

library global_lib;
use global_lib.numeric_std.all;
use global_lib.global.all;

library vci_lib;
use vci_lib.vci_pack.all;

entity vci_split_sim is

  generic(n_t    : positive := 100000;      -- Number of test vectors
          rnd_ce : boolean  := false;
          debug  : boolean  := false);
  port(p_vci_initiator_out   : out vci_t2i;
       p_vci_target_sec_out  : out vci_i2t;
       p_vci_target_unsec_out: out vci_i2t);

end entity vci_split_sim;

architecture sim of vci_split_sim is

  signal clk, srstn, ce, en: std_ulogic;
  signal eos: boolean;
  signal p_vci_initiator_req, p_vci_target_sec_req, p_vci_target_unsec_req: vci_i2t;
  signal p_vci_initiator_rsp, p_vci_target_sec_rsp, p_vci_target_unsec_rsp: vci_t2i;
  signal p_addr, p_size: addr_t;
  signal sec_unsec: boolean;
begin

  p_vci_initiator_out    <= p_vci_initiator_rsp;
  p_vci_target_sec_out   <= p_vci_target_sec_req;
  p_vci_target_unsec_out <= p_vci_target_unsec_req;

  process
  begin
    clk <= '0';
    if rnd_ce then
      ce <= std_ulogic_rnd;
    else
      ce <= '1';
    end if;
    wait for 1 ns;
    clk <= '1';
    wait for 1 ns;
    if eos then
      wait;
    end if;
  end process;

  process
    variable l: line;
  begin
    eos <= false;
    srstn <= '0';
    for i in 1 to 10 loop
      wait until rising_edge(clk) and ce = '1';
    end loop;
    srstn <= '1';
    for i in 1 to n_t loop
    if (i * 10) mod n_t = 0 then
        write(l, i);
        write(l, string'("/"));
        write(l, n_t);
        writeline(output, l);
    end if;
      en <= std_ulogic_rnd;
      p_addr <= std_ulogic_vector_rnd(ADDR_SIZE);
      p_size <= std_ulogic_vector_rnd(ADDR_SIZE);
      wait until rising_edge(clk) and ce='1' and ((p_vci_target_sec_rsp.rspval='1' and p_vci_target_sec_req.rspack='1') or (p_vci_target_unsec_rsp.rspval='1' and p_vci_target_unsec_req.rspack='1')); -- either the secured or the unsecured recognizes the request as served
    end loop;
    eos <= true;
    write(l, string'("regression test passed"));
    writeline(output, l);
    wait;
  end process;

  irnd_vci_target_sec: entity vci_lib.rnd_vci_target(rtl)
  port map( clk => clk,
            srstn => srstn,
            ce => ce,
            eos => eos,
            p_vci_target_req => p_vci_target_sec_req,
            p_vci_target_rsp => p_vci_target_sec_rsp);

  irnd_vci_target_unsec: entity vci_lib.rnd_vci_target(rtl)
  port map( clk => clk,
            srstn => srstn,
            ce => ce,
            eos => eos,
            p_vci_target_req => p_vci_target_unsec_req,
            p_vci_target_rsp => p_vci_target_unsec_rsp);

  irnd_vci_initiator: entity vci_lib.rnd_vci_initiator(rtl)  -- random VCI Initiator
  port map( clk => clk,
            srstn => srstn,
            ce => ce,
            eos => eos,
            p_vci_initiator_rsp => p_vci_initiator_rsp,
            p_vci_initiator_req => p_vci_initiator_req);

  ivci_split: entity work.vci_split(rtl)
  port map(clk => clk,
           srstn => srstn,
           ce => ce,
           p_vci_ini_in        => p_vci_initiator_req,
           p_vci_ini_out       => p_vci_initiator_rsp,
           p_vci_tgt_sec_in    => p_vci_target_sec_rsp,
           p_vci_tgt_sec_out   => p_vci_target_sec_req,
           p_vci_tgt_uns_in    => p_vci_target_unsec_rsp,
           p_vci_tgt_uns_out   => p_vci_target_unsec_req,
           p_enable            => en,
           p_addr              => p_addr,
           p_size              => p_size);

  target_checker: process
  begin
    -- IDLE
    wait until rising_edge(clk) and ce='1' and p_vci_initiator_req.cmdval='1';
    if inclusion(p_vci_initiator_req.addr, p_addr, std_ulogic_vector(unsigned(p_addr) + unsigned(p_size))) = '1' and en='1' then
        sec_unsec <= true;
    else
        sec_unsec <= false;
    end if;
    -- CMD
    wait until rising_edge(clk) and ce='1' and p_vci_initiator_req.eop='1' and p_vci_initiator_req.cmdval='1' and (p_vci_target_sec_rsp.cmdack='1' or p_vci_target_unsec_rsp.cmdack='1');
    if sec_unsec then
      assert p_vci_target_sec_rsp.cmdack='1'
        report "Wrong target, it should be the protected one"
        severity failure;
    else
      assert p_vci_target_unsec_rsp.cmdack='1'
        report "Wrong target, it should be the unprotected one"
        severity failure;
    end if;
    -- RSP
    wait until rising_edge(clk) and ce='1' and ((p_vci_target_sec_rsp.reop='1' and p_vci_target_sec_rsp.rspval='1') or (p_vci_target_unsec_rsp.reop='1' and p_vci_target_unsec_rsp.rspval='1')) and p_vci_initiator_req.rspack='1';
    if sec_unsec then
      assert p_vci_target_sec_rsp.reop='1' and p_vci_target_sec_rsp.rspval='1'
        report "Wrong target, it should be the protected one"
        severity failure;
    else
      assert p_vci_target_unsec_rsp.reop='1' and p_vci_target_unsec_rsp.rspval='1'
        report "Wrong target, it should be the unprotected one"
        severity failure;
    end if;
  end process;

end architecture sim;
